/*
    This library function uses the Berger and Rigoutsos clustering
    algorithm to create the adaptive mesh rectangular blocks.  I
    used the references:

       http://www-lmc.imag.fr/IDOPT/AGRIF/documentation/doc_AGRIF3DUser/node9.html

    and the paper:

       Gunney, B. T. N. and A. M. Wissink, A Task-parallel Clustering Algorithm for
          Structured AMR, 2004.

    which I got at:

       www.llnl.gov/tid/lof/documents/pdf/313502.pdf

                     Updated 2/30/07

    SLFCFD source file
    Version:  1.0
    Copyright (C) 2007  San Le

    The source code contained in this file is released under the
    terms of the GNU Library General Public License.

*/

#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "../compressible/cfdconst.h"
#include "../compressible/cfdstruct.h"
#include "amr.h"

int numnp, sof, nx, ny, nz;
double dx, dy, dz;
int flag_2D, flag_3D;
double tolerance;

int cluster ( INC_NODE *node, SUBGRID *subgrid, int *zero_cross_i, int *zero_cross_j)
{
	int i, i2, j, j2;
	int min_i, max_i, min_j, max_j, grid_size, node_num, node_counter;
	int dum, dum0, dum1, dum2, dum3, dum4, dum5, dum6, dum7, dum8, dum9,
	    dum10, dum11, dum12, dum13, dum14, dum15, dum16, dum17, dum18, dum19,
	    dum20, dum21, dum22, dum23, dum24;
	double fdum, fdum2, fdum3;
	int sig_i[200], sig_j[200], temp_sig_i[200], temp_sig_j[200],
		lap_sig_i[200], lap_sig_j[200];

	memset(sig_i,0,200*sizeof(int));
	memset(sig_j,0,200*sizeof(int));
	memset(temp_sig_i,0,200*sizeof(int));
	memset(temp_sig_j,0,200*sizeof(int));
	memset(lap_sig_i,0,200*sizeof(int));
	memset(lap_sig_j,0,200*sizeof(int));

	min_i = nx-1; max_i = 0;
	min_j = ny-1; max_j = 0;

	node_counter = 0;
	for( i2 = subgrid[0].istart; i2 <= subgrid[0].iend; ++i2)
	{
	    for( j2 = subgrid[0].jstart; j2 <= subgrid[0].jend; ++j2)
	    {

		node_num = ny*i2 + j2;
		if( node[node_num].refine_flag )
		{
		    if( min_i > node[node_num].i) min_i = node[node_num].i;
		    if( max_i < node[node_num].i) max_i = node[node_num].i;

		    if( max_j < node[node_num].j) max_j = node[node_num].j;
		    if( min_j > node[node_num].j) min_j = node[node_num].j;

/* Calculate the signatures */

		    temp_sig_i[i2] += 1;
		    temp_sig_j[j2] += 1;
		    ++node_counter;
		}
	    }
	}

	subgrid[0].istart = min_i;
	subgrid[0].iend = max_i;
	subgrid[0].jstart = min_j;
	subgrid[0].jend = max_j;
	
	dum1 = max_i - min_i + 1;
	dum2 = max_j - min_j + 1;
	grid_size = dum1*dum2;

	if(  grid_size < 20 )
	{
		subgrid[0].stop_flag = 1;
		return 1;
	}

	fdum = (double)node_counter/((double)grid_size);

	if( fdum > 0.8 )
	{
		subgrid[0].stop_flag = 1;
		return 1;
	}

	for( i2 = 0; i2 <= max_i - min_i; ++i2)
	{
		sig_i[i2] = temp_sig_i[i2 + min_i];
	}
	for( i2 = 0; i2 <= max_j - min_j; ++i2)
	{
		sig_j[i2] = temp_sig_j[i2 + min_j];
	}

	subgrid[0].istart = min_i; subgrid[0].iend = max_i;
	subgrid[0].jstart = min_j; subgrid[0].jend = max_j;

/* Calculate the Laplace signatures */

	*zero_cross_i = min_i;
	*zero_cross_j = min_j;

	dum3 = 0;
	dum4 = 0;
	lap_sig_i[0] = sig_i[1]- 2*sig_i[0];
	lap_sig_i[1] = sig_i[2] - 2.0*sig_i[1] + sig_i[0];
	for( i2 = 2; i2 < max_i - min_i; ++i2)
	{
	    lap_sig_i[i2] = sig_i[i2+1] - 2.0*sig_i[i2] + sig_i[i2-1];
	    dum3 = abs(lap_sig_i[i2] - lap_sig_i[i2-1]);
	    if( dum4 < dum3 )
	    {
		dum4 = dum3;
		*zero_cross_i = i2 + min_i;
	    }
	    /*printf("qqqq %7d %7d %7d %7d %7d %7d\n",
		i2, i2 + min_i, lap_sig_i[i2], sig_i[i2], dum3, dum4);*/
	}

	dum3 = 0;
	dum5 = 0;
	lap_sig_j[0] = sig_j[1] - 2*sig_j[0];
	lap_sig_j[1] = sig_j[2] - 2.0*sig_j[1] + sig_j[0];
	for( i2 = 2; i2 < max_j - min_j; ++i2)
	{
	    lap_sig_j[i2] = sig_j[i2+1] - 2.0*sig_j[i2] + sig_j[i2-1];
	    dum3 = abs(lap_sig_j[i2] - lap_sig_j[i2-1]);
	    if( dum5 < dum3 )
	    {
		dum5 = dum3;
		*zero_cross_j = i2 + min_j;
	    }
	    /*printf("ssss %7d %7d %7d %7d %7d %7d\n",
		i2, i2 + min_j, lap_sig_j[i2], sig_j[i2], dum3, dum5);*/
	}

	if( *zero_cross_i <= min_i + 1 && *zero_cross_j <= min_j + 1)
	{
		subgrid[0].stop_flag = 1;
		return 1;
	}

	if(dum4 > dum5) *zero_cross_j = subgrid[0].jend;   /* Divide along j so keep i the same as before. */
	else *zero_cross_i = subgrid[0].iend;

	/*printf("ffff %7d %7d %7d %7d %7d %7d %7d\n",
		dum4, dum5, *zero_cross_i, *zero_cross_j, subgrid[0].iend,
		subgrid[0].jend, subgrid[0].stop_flag);*/


	return 1;
}
