/* The functions in this file do the regridding from the coarse grid to
   the fine grid and from the fine grid to the coarse after calculation on
   the refined grid is made.

   There is also a function which sets the parameters of the subgrid.

                        Updated 4/9/07

    SLFCFD source file
    Version:  1.0
    Copyright (C) 2007  San Le

    The source code contained in this file is released under the
    terms of the GNU Library General Public License.
 */

#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "../compressible/cfdconst.h"
#include "../compressible/cfdstruct.h"
#include "amr.h"

extern double dx, dy, dt;

int setSubgrid( SUBGRID *subgrid_f )
{
/* This code does 2 things:

   1) Sets the value of dx, dy, and dt for the subgrid_f.

   2) Sets the dimensions of the subgrid_f which doubles the number of points in i and j
      as the original grid it is in.

                        Updated 3/28/07
*/

	int i, i2, j, j2, node_num, dum, nx_fine, ny_fine, ny_coarse;
	int ce, ce2, xm1, xm2, xp1, ym1, ym2, yp1, yp2, xp1yp1;

	subgrid_f[0].nx = 2.0*(subgrid_f[0].iend - subgrid_f[0].istart) + 1;
	subgrid_f[0].ny = 2.0*(subgrid_f[0].jend - subgrid_f[0].jstart) + 1;

	switch(subgrid_f[0].refine_num)
	{
	    case 1: /* Divide grid discretization by 2 */
		subgrid_f[0].dx = dx/2.0;
		subgrid_f[0].dy = dy/2.0;
		subgrid_f[0].dt = dt/2.0;
	    break;
	    case 2: /* Divide grid discretization by 4 */
		subgrid_f[0].dx = dx/4.0;
		subgrid_f[0].dy = dy/4.0;
		subgrid_f[0].dt = dt/4.0;
	    break;
	    case 3: /* Divide grid discretization by 8 */
		subgrid_f[0].dx = dx/8.0;
		subgrid_f[0].dy = dy/8.0;
		subgrid_f[0].dt = dt/8.0;
	    break;
	    default:
		subgrid_f[0].dx = dx;
		subgrid_f[0].dy = dy;
		subgrid_f[0].dt = dt;
	    break;
	}

	return 1;
}

int refineGridSet( double *vec_coarse, double *vec_fine, SUBGRID subgrid_c, SUBGRID subgrid_f )
{
/* This program prepares the subgrid.  The resulting subgrid may be a
   subgrid of the intial mesh or a subgrid of a subgrid.  This code does
   essentially 2 things:

   1) It takes the patch of grid and doubles the number nodal points in
      the i and j direction. 

   2) The nodal values from the coarser grid are assiged to the corresponding
      points of the subgrid.  Then the points which fall in between the nodes
      of the coarser grid have their values interpolted from the original.

                        Updated 4/9/07
*/

	int i, i2, j, j2, node_num, dum, nx_fine, ny_fine, ny_coarse;
	int ce, ce2, xm1, xm2, xp1, ym1, ym2, yp1, yp2, xp1yp1;

	nx_fine = subgrid_f.nx;
	ny_fine = subgrid_f.ny;

	ny_coarse = subgrid_c.ny;

	j2 = 0;
	for( j = subgrid_f.jstart; j < subgrid_f.jend + 1; ++j)
	{
	    i2 = 0;
	    for( i = subgrid_f.istart; i < subgrid_f.iend + 1; ++i)
	    {

/* For the fine nodes which directly fall on the coarse nodes (i,j) */

		ce = ny_coarse*i + j;
		ce = ny_coarse*i + j;
		xp1 = ny_coarse*(i+1) + j;
		yp1 = ny_coarse*i + j+1;
		xp1yp1 = ny_coarse*(i+1) + j+1;

		ce2 = ny_fine*i2 + j2;
		vec_fine[ce2] = vec_coarse[ce];
		/*printf("%7d %7d %7d %7d %7d %7d\n", i, j, i2, j2, ce, ce2);*/
		/*printf("mmmm %7d %7d %7d %7d %16.8e %16.8e\n", i, j, i2, j2,
			vec_fine[ce2], vec_coarse[ce]);*/

		if( i < subgrid_f.iend )
		{
/* For the fine nodes which directly fall between the coarse nodes at (i,j) and (i+1,j) */

		    ce2 = ny_fine*(i2+1) + j2;
		    vec_fine[ce2] = .5*(vec_coarse[ce] + vec_coarse[xp1]);
		    /*printf("nnnn %7d %7d %7d %7d %16.8e %16.8e %16.8e\n",
			i, j, i2, j2, vec_fine[ce2], vec_coarse[ce], vec_coarse[xp1]);*/
		}
		if( j < subgrid_f.jend )
		{
/* For the fine nodes which directly fall between the coarse nodes at (i,j) and (i,j+1) */

		    ce2 = ny_fine*i2 + j2+1;
		    vec_fine[ce2] = .5*(vec_coarse[ce] + vec_coarse[yp1]);
		    /*printf("oooo %7d %7d %7d %7d %16.8e %16.8e %16.8e\n",
			i, j, i2, j2, vec_fine[ce2], vec_coarse[ce], vec_coarse[yp1]);*/
		}

		if( i < subgrid_f.iend && j < subgrid_f.jend )
		{
/* For the fine nodes which directly fall between the coarse nodes at (i,j), (i+1,j), (i,j+1),
   and (i+1,j+1) */

		    ce2 = ny_fine*(i2+1) + j2+1;
		    vec_fine[ce2] = .25*(vec_coarse[ce] + vec_coarse[xp1] + vec_coarse[yp1] + vec_coarse[xp1yp1]);
		}

		i2 += 2;
	    }
	    j2 += 2;
	}

/* The code below does the same as that above but the above is more readable. */

#if 0
	j2 = 0;
	for( j = subgrid_f.jstart; j < subgrid_f.jend; ++j)
	{
	    i2 = 0;
	    for( i = subgrid_f.istart; i < subgrid_f.iend; ++i)
	    {

/* For the fine nodes which directly fall on the coarse nodes (i,j) */

		ce = ny_coarse*i + j;
		ce2 = ny_fine*i2 + j2;
		vec_fine[ce2] = vec_coarse[ce];
		/*printf("%7d %7d %7d %7d %7d %7d\n", i, j, i2, j2, ce, ce2);*/ 

/* For the fine nodes which directly fall between the coarse nodes at (i,j) and (i+1,j) */

		xp1 = ny_coarse*(i+1) + j;
		ce2 = ny_fine*(i2+1) + j2;
		vec_fine[ce2] = .5*(vec_coarse[ce] + vec_coarse[xp1]);

/* For the fine nodes which directly fall between the coarse nodes at (i,j) and (i,j+1) */

		yp1 = ny_coarse*i + j+1;
		ce2 = ny_fine*i2 + j2+1; 
		vec_fine[ce2] = .5*(vec_coarse[ce] + vec_coarse[yp1]);

/* For the fine nodes which directly fall between the coarse nodes at (i,j), (i+1,j), (i,j+1),
   and (i+1,j+1) */

		xp1 = ny_coarse*(i+1) + j;
		yp1 = ny_coarse*i + j+1;
		xp1yp1 = ny_coarse*(i+1) + j+1;
		ce2 = ny_fine*(i2+1) + j2+1; 
		vec_fine[ce2] = .25*(vec_coarse[ce] + vec_coarse[xp1] + vec_coarse[yp1] + vec_coarse[xp1yp1]);

		i2 += 2;
	    }
	    j2 += 2;
	}

 /* For nodes at the end, j = subgrid_f.jend. */

	i2 = 0;
	for( i = subgrid_f.istart; i < subgrid_f.iend; ++i)
	{
		j = subgrid_f.jend;
		j2 = subgrid_f.ny - 1;

/* For the fine nodes which directly fall on the coarse nodes (i,j). */

		ce = ny_coarse*i + j;
		ce2 = ny_fine*i2 + j2;
		vec_fine[ce2] = vec_coarse[ce];
		/*printf("%7d %7d %7d %7d %7d %7d\n", i, j, i2, j2, ce, ce2);*/ 

	/*printf("%7d %7d %7d %7d\n",
		subgrid_f.iend - subgrid_f.istart, subgrid_f.jend - subgrid_f.jstart, i2, j2); */

/* For the fine nodes which directly fall between the coarse nodes at (i,j) and (i+1,j) */

		xp1 = ny_coarse*(i+1) + j;
		ce2 = ny_fine*(i2+1) + j2;
		vec_fine[ce2] = .5*(vec_coarse[ce] + vec_coarse[xp1]);

		i2 += 2;
	}

/* For nodes at the end, i = subgrid_f.iend. */

	j2 = 0;
	for( j = subgrid_f.jstart; j < subgrid_f.jend; ++j)
	{
		i = subgrid_f.iend;
		i2 = subgrid_f.nx - 1;

/* For the fine nodes which directly fall on the coarse nodes (i,j). */

		ce = ny_coarse*i + j;
		ce2 = ny_fine*i2 + j2;
		vec_fine[ce2] = vec_coarse[ce];
		/*printf("%7d %7d %7d %7d %7d %7d\n", i, j, i2, j2, ce, ce2);*/ 

/* For the fine nodes which directly fall between the coarse nodes at (i,j) and (i,j+1) */

		yp1 = ny_coarse*i + j+1;
		ce2 = ny_fine*i2 + j2+1; 
		vec_fine[ce2] = .5*(vec_coarse[ce] + vec_coarse[yp1]);

		j2 += 2;
	}

/* For the corner node, i = subgrid_f.iend, j = subgrid_f.jend. */

	i = subgrid_f.iend;
	i2 = subgrid_f.nx - 1;
	j = subgrid_f.jend;
	j2 = subgrid_f.ny - 1;
	ce = ny_coarse*i + j;
	ce2 = ny_fine*i2 + j2;
	vec_fine[ce2] = vec_coarse[ce];
	/*printf("%7d %7d %7d %7d %7d %7d\n", i, j, i2, j2, ce, ce2);*/ 
#endif

	return 1;
}


int coarseGridSet( double *vec_coarse, double *vec_fine, SUBGRID subgrid_c, SUBGRID subgrid_f )
{
/* This program prepares takes the data values from the fine grid and
   puts it on the corresponding ponits of the coarse grid. 

                        Updated 4/9/07
*/

	int i, i2, j, j2, node_num, dum, nx_fine, ny_fine, ny_coarse;
	int ce, ce2, xm1, xm2, xp1, ym1, ym2, yp1, yp2, xp1yp1;

	nx_fine = subgrid_f.nx;
	ny_fine = subgrid_f.ny;

	ny_coarse = subgrid_c.ny;

	j2 = 0;
	for( j = subgrid_f.jstart; j < subgrid_f.jend + 1; ++j)
	{
	    i2 = 0;
	    for( i = subgrid_f.istart; i < subgrid_f.iend + 1; ++i)
	    {

/* For the fine nodes which directly fall on the coarse nodes (i,j) */

		ce = ny_coarse*i + j;
		ce2 = ny_fine*i2 + j2;
		vec_coarse[ce] = vec_fine[ce2];
		/*printf("%7d %7d %7d %7d %7d %7d\n", i, j, i2, j2, ce, ce2);*/ 
		i2 += 2;
	    }
	    j2 += 2;
	}

	return 1;
}

int interp( double *vec_coarse, double *vec_fine, SUBGRID subgrid_c, SUBGRID subgrid_f )
{
/* This is a coarse to fine grid interpolation.  It is based on the code from page 880 of:

   Press, William H., et. al., Numerical Recipes in C 2nd Ed., Cambridge University
        Press, New York, 1992.

   I got the same results as with my own code.  So the problem is with the interpolation.

                        Updated 4/9/07
*/

	int i, i2, j, j2, node_num, dum, nx_fine, ny_fine, ny_coarse;
	int ce, ce2, xm1, xm2, xp1, ym1, ym2, yp1, yp2, xp1yp1;
	int ic, iif, jc, jf, ncy, nfx, nfy;

	nfx = subgrid_f.nx;
	nfy = subgrid_f.ny;

	ncy = subgrid_c.ny;

	jf = 0;
	for( jc = subgrid_f.jstart; jc < subgrid_f.jend + 1; ++jc)
	{
	    iif = 0;
	    for( ic = subgrid_f.istart; ic < subgrid_f.iend + 1; ++ic)
	    {

/* For the fine nodes which directly fall on the coarse nodes (ic,jc) */

		vec_fine[nfy*iif + jf] = vec_coarse[ncy*ic + jc];
		/*printf("mmmm %7d %7d %7d %7d %16.8e %16.8e\n",
			ic, jc, iif, jf, vec_fine[nfy*iif + jf], vec_coarse[ncy*ic + jc]);*/
		iif += 2;
	    }
	    jf += 2;
	}

/* For the fine nodes which fall on the odd number columns. */

	for( jf = 0; jf < nfy; jf+=2)
	{
	    for( iif = 1; iif < nfx; iif+=2)
	    {
		vec_fine[nfy*iif + jf] = 0.5*(vec_fine[nfy*(iif+1) + jf] + vec_fine[nfy*(iif-1) + jf]);
		/*printf("nnnn %7d %7d %7d %7d %16.8e %16.8e %16.8e\n",
			ic, jc, iif, jf, vec_fine[nfy*iif + jf], vec_fine[nfy*(iif+1) + jf], vec_fine[nfy*(iif-1) + jf]);*/
	    }
	}

/* For the fine nodes which fall on the even number columns. */

	for( jf = 1; jf < nfy; jf+=2)
	{
	    for( iif = 0; iif < nfx; ++iif)
	    {
		vec_fine[nfy*iif + jf] = 0.5*(vec_fine[nfy*iif + jf+1] + vec_fine[nfy*iif + jf-1]);
		/*printf("oooo %7d %7d %7d %7d %16.8e %16.8e %16.8e\n",
			ic, jc, iif, jf, vec_fine[nfy*iif + jf], vec_fine[nfy*iif + jf+1], vec_fine[nfy*iif + jf-1]);*/
	    }
	}

	return 1;
}

int refineGridSet2( double *vec_coarse, double *vec_fine, SUBGRID subgrid_c, SUBGRID subgrid_f )
{
/* This program prepares the subgrid.  The resulting subgrid may be a
   subgrid of the intial mesh or a subgrid of a subgrid.  This code does
   essentially 2 things:

   1) It takes the patch of grid and doubles the number nodal points in
      the i and j direction. 

   2) The nodal values from the coarser grid are assiged to the corresponding
      points of the subgrid.  Then the points which fall in between the nodes
      of the coarser grid have their values interpolted from the original.

                        Updated 4/9/07
*/

	int i, i2, j, j2, node_num, dum, nx_fine, ny_fine, ny_coarse;
	int ce, ce2, xm1, xm2, xp1, ym1, ym2, yp1, yp2, xp1yp1, xp1ym1;

	nx_fine = subgrid_f.nx;
	ny_fine = subgrid_f.ny;

	ny_coarse = subgrid_c.ny;

	j2 = 0;
	for( j = subgrid_f.jstart; j < subgrid_f.jend + 1; ++j)
	{
	    i2 = 0;
	    for( i = subgrid_f.istart; i < subgrid_f.iend + 1; ++i)
	    {

/* For the fine nodes which directly fall on the coarse nodes (i,j) */

		ce = ny_coarse*i + j;
		xp1 = ny_coarse*(i+1) + j;
		xm1 = ny_coarse*(i-1) + j;
		yp1 = ny_coarse*i + j+1;
		ym1 = ny_coarse*i + j-1;
		xp1yp1 = ny_coarse*(i+1) + j+1;
		xp1ym1 = ny_coarse*(i+1) + j-1;

		ce2 = ny_fine*i2 + j2;
		vec_fine[ce2] = vec_coarse[ce];
		/*printf("%7d %7d %7d %7d %7d %7d\n", i, j, i2, j2, ce, ce2);*/

		if( i < subgrid_f.iend )
		{
/* For the fine nodes which directly fall between the coarse nodes at (i,j) and (i+1,j) */

		    ce2 = ny_fine*(i2+1) + j2;
		    vec_fine[ce2] = .25*(vec_coarse[ce] + vec_coarse[xp1]) +
			.125*(vec_coarse[yp1] + vec_coarse[ym1] + vec_coarse[xp1yp1] + vec_coarse[xp1ym1]);
		}
		if( j < subgrid_f.jend )
		{
/* For the fine nodes which directly fall between the coarse nodes at (i,j) and (i,j+1) */

		    ce2 = ny_fine*i2 + j2+1;
		    vec_fine[ce2] = .25*(vec_coarse[ce] + vec_coarse[yp1]) + 
			.125*(vec_coarse[xp1] + vec_coarse[xm1] + vec_coarse[xp1yp1] + vec_coarse[xp1ym1]);
		}

		if( i < subgrid_f.iend && j < subgrid_f.jend )
		{
/* For the fine nodes which directly fall between the coarse nodes at (i,j), (i+1,j), (i,j+1),
   and (i+1,j+1) */

		    ce2 = ny_fine*(i2+1) + j2+1;
		    vec_fine[ce2] = .25*(vec_coarse[ce] + vec_coarse[xp1] + vec_coarse[yp1] + vec_coarse[xp1yp1]);
		}

		i2 += 2;
	    }
	    j2 += 2;
	}

	return 1;
}
