/*
    This utility function uses the conjugate gradient method
    to solve the 2-D Poisson Equation for a fluid dynamics problem.
    The system has the form [A][U] = [B].  The first function
    assembles the P matrix.  It is called by the second function
    which allocates the memory and goes through the steps of
    the algorithm.

	          Last Update 3/17/07

    SLFCFD source file
    Version:  1.4
    Copyright (C) 2007  San Le

    The source code contained in this file is released under the
    terms of the GNU Library General Public License.
 
*/

#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "../AMR/amr.h"

#define SMALL      1.e-20

extern int sof;

extern double dx, dy, dz;
extern double dt, err, Re;

int Perimeter_Boundary ( double *, SUBGRID );

int Coarse_Boundary ( double *, SUBGRID, SUBGRID );

int dotX(double *, double *, double *, int);

int PoisConjPassemble_AMR( double *P_global_CG, double *U, int dof, int block_size)
{
/* This function assembles the P_global_CG matrix for the displacement calculation by
   taking the product [K_el]*[U_el].  Some of the [K_el] is stored in [A].

                  Updated 2/9/07
*/
	int i, i1, i2, j, k;
	int check, dum, node, num_block;
	double fdum, fdum2;

	memset(P_global_CG,0,dof*sof);

/* Because of the discretized Poisson Equation has a standard form:

        http://en.wikipedia.org/wiki/Discrete_Poisson_equation

   I can optimize these multiplications by only taking the product of the nonzero terms.
   So I break up the multiplication into 3 parts.
*/

	num_block = dof/block_size; /* Obviously, dof has to be a multiple of block_size. */

/* Part 1.  This is for the first block of the block tridiagonal */

	*(P_global_CG) = 4.0*(*(U)) - *(U+1) - *(U + block_size);

	for( i = 1; i < block_size - 1; ++i )
	{
		*(P_global_CG + i) =
			-*(U-1 + i) + 4.0*(*(U + i)) - *(U+1 + i) - *(U + block_size + i); 
	}

	i = block_size - 1;
	*(P_global_CG + i) = -*(U-1 + i) + 4.0*(*(U + i)) - *(U + block_size + i); 

/* Part 2.  This is for the middle blocks of the block tridiagonal */

	for( k = 1; k < num_block - 1; ++k )
	{
	    dum = block_size*k;
	    *(P_global_CG + dum) =
		-*(U + dum - block_size)  + 4.0*(*(U + dum)) - *(U+1 + dum) - *(U + block_size + dum); 

	    for( i = 1; i < block_size - 1; ++i )
	    {
		dum = block_size*k + i;
		*(P_global_CG + dum) =
		    -*(U + dum - block_size) -*(U-1 + dum) + 4.0*(*(U + dum)) - *(U+1 + dum) - *(U + block_size + dum); 
	    }

	    i = block_size - 1;
	    dum = block_size*k + i;
	    *(P_global_CG + dum) =
		-*(U + dum - block_size) -*(U-1 + dum) + 4.0*(*(U + dum)) - *(U + block_size + dum); 
	}

/* Part 3.  This is for the last block of the block tridiagonal */

	i = dof - block_size;
	*(P_global_CG + i) = -*(U + i - block_size) + 4.0*(*(U + i)) - *(U+1 + i);

	for( i = dof - block_size + 1; i < dof - 1; ++i )
	{
		*(P_global_CG + i) =
		    -*(U + i - block_size) -*(U-1 + i) + 4.0*(*(U + i)) - *(U+1 + i);
	}

	i = dof - 1;
	*(P_global_CG + i) = -*(U + i - block_size) -*(U-1 + i) + 4.0*(*(U + i));

	return 1;
}


int PoisConjGrad_AMR( double *X, double *Y, double *prs, SUBGRID subgrid_f)
{
/* This function does memory allocation and uses the conjugate gradient
   method to solve the linear system which arises from the discrete
   Poisson equation.

                  Last Update 3/17/07

   It is taken from the algorithm 10.3.1 given in "Matrix Computations",
   by Golub, page 534.
*/
	int i, j, k, sofmf, ptr_inc, node_num, dof, block_size;
	int check, counter;
	double *mem_double;
	double *p, *P_global_CG, *r, *z;
	double *pre_cond, *B, *U;
	double alpha, alpha2, beta;
	double fdum, fdum2, fdum3;
	double dx_f, dy_f, dt_f;
	int ce, xm1, xm2, xp1, ym1, ym2, yp1, yp2, xm1ym1, xm1yp1, xp1ym1, xp1yp1;
	int bce, bxm1, bxm2, bxm3, bxp1, bxp2, bxp3, bym1, bym2, byp1, byp2,
		bxm1ym1, bxm1yp1, bxp1ym1, bxp1yp1;
	double pxm1, pxp1, pym1, pyp1;
	int nx_f, ny_f;
	int  iteration_max, iteration_const, iteration;
	double tolerance;

	iteration_max = 2000;
	iteration = 0;
	tolerance = 1.0e-13;

	nx_f = subgrid_f.nx;
	ny_f = subgrid_f.ny;

	dof = nx_f*ny_f;
	block_size = nx_f;

/* For the doubles */
	sofmf = 7*dof;
	mem_double=(double *)calloc(sofmf,sizeof(double));

	if(!mem_double )
	{
		printf( "failed to allocate memory for doubles\n ");
		exit(1);
	}

/* For the Conjugate Gradient Method doubles */

	                                        ptr_inc = 0;
	p=(mem_double+ptr_inc);                 ptr_inc += dof;
	P_global_CG=(mem_double+ptr_inc);       ptr_inc += dof;
	r=(mem_double+ptr_inc);                 ptr_inc += dof;
	z=(mem_double+ptr_inc);                 ptr_inc += dof;
	B=(mem_double+ptr_inc);                 ptr_inc += dof;
	U=(mem_double+ptr_inc);                 ptr_inc += dof;
	pre_cond=(mem_double+ptr_inc);          ptr_inc += dof;

/* Using Conjugate gradient method to find displacements */

	memset(P_global_CG,0,dof*sof);
	memset(B,0,dof*sof);
	memset(p,0,dof*sof);
	memset(r,0,dof*sof);
	memset(z,0,dof*sof);

/* Create the RHS vector

IMPORTANT!!!

I needed to reorder the indices such that the varaibles which are used
for this calculation, B, U, P_global_CG, p, r, and z follow natural ordering.
This is why I have both "ce" and "bce" below.  So any time I am dealing
with theses variables, there should be a "b" prefix to the indexing quantity
(e.g.  bxm1, bxm2, byp1, byp2).  Remember this when dealing with these
variables and note especially that "ny_f*i + j" becomes "i + nx_f*j".
*/

	dx_f = subgrid_f.dx;
	dy_f = subgrid_f.dy;
	dt_f = subgrid_f.dt;
	fdum = dx_f/dt_f;

	for( j = 0; j < ny_f; ++j )
	{
	    for( i = 0; i < nx_f; ++i )
	    {
		   bce  = i + nx_f*j;
		   ce  = ny_f*i + j;
		   xm1 = ny_f*(i-1) + j;
		   xp1 = ny_f*(i+1) + j;
		   ym1 = ny_f*i + (j-1);
		   yp1 = ny_f*i + (j+1);

		   *(B + bce) = -fdum*(*(X + ce) - *(X + xm1) + *(Y + ce) - *(Y + ym1));
		   *(U + bce) = *(prs + ce);
	    }
	}

/* The initial setting of "r" is for it to be set to "B" above. */

	/*memcpy( r, B, dof*sizeof(double));*/

/* Use the value of the last pressure as an initial guess. */

	check = PoisConjPassemble_AMR( P_global_CG, U, dof, block_size);
	if(!check) printf( " Problems with PoisConjPassemble_AMR \n");

	for( i = 0; i < dof; ++i )
	{
		*(r + i) =  *(B + i) - *(P_global_CG + i);
	}

	for( k = 0; k < 1; ++k )
	{
	    for( j = 0; j < dof; ++j )
	    {
		*(pre_cond + j) += 1.0 + SMALL;
		*(z + j) = *(r + j)/(*(pre_cond + j));
		*(p+j) = *(z+j);
	    }

	    check = Perimeter_Boundary( p, subgrid_f);
	    if(!check) printf( " Problems with Perimeter_Boundary\n");

	    check = Perimeter_Boundary( r, subgrid_f);
	    if(!check) printf( " Problems with Perimeter_Boundary\n");

	    alpha = 0.0;
	    alpha2 = 0.0;
	    beta = 0.0;
	    fdum2 = 1000.0;
	    counter = 0;
	    check = dotX(&fdum, r, z, dof);

	    printf("\n iteration %3d iteration max %3d \n", iteration, iteration_max);
	    /*for( iteration = 0; iteration < iteration_max; ++iteration )*/

/* The outer "k" loop is for when the newly calculated pressures are extrapolated to the
   boundaries.  I start the calculation over until the code converges.
*/
	    while( fabs(fdum2) > tolerance && counter < iteration_max )
	    {

		/*printf( "\n %3d %16.8e\n",counter, fdum2);*/
		check = PoisConjPassemble_AMR( P_global_CG, p, dof, block_size);
		if(!check) printf( " Problems with PoisConjPassemble_AMR \n");

		check = Perimeter_Boundary( P_global_CG, subgrid_f);
		if(!check) printf( " Problems with Perimeter_Boundary\n");

		check = dotX(&alpha2, p, P_global_CG, dof);	
		alpha = fdum/(SMALL + alpha2);
		for( j = 0; j < dof; ++j )
		{
		    *(U+j) += alpha*(*(p+j));
		    *(r+j) -= alpha*(*(P_global_CG+j));
		    *(z + j) = *(r + j)/(*(pre_cond + j));
		}

		check = dotX(&fdum2, r, z, dof);
		beta = fdum2/(SMALL + fdum);
		fdum = fdum2;
		
		for( j = 0; j < dof; ++j )
		{
		    /*printf("\n  %3d %12.7f  %14.5f ",j,*(U+j),*(P_global_CG+j));*/
		    /*printf( "%4d %14.5f  %14.5f  %14.5f  %14.5f %14.5f\n",j,alpha,
			*(U+j),*(r+j),*(P_global_CG+j),*(B+j));
		    printf( "%4d %14.8f  %14.8f  %14.8f  %14.8f %14.8f\n",j,
			*(U+j)*beta,*(r+j)*beta,*(P_global_CG+j)*alpha,
			*(B+j)*alpha);*/
		    *(p+j) = *(z+j)+beta*(*(p+j));
		}

		check = Perimeter_Boundary( p, subgrid_f);
		if(!check) printf( " Problems with Perimeter_Boundary\n");

		++counter;
	    }
	}


	if(counter > iteration_max - 1 )
	{
		printf( "\nMaximum iterations %4d reached.  Residual is: %16.8e\n",counter,fdum2);
		printf( "Problem may not have converged during Conj. Grad.\n");
	}

/*
The lines below are for testing the quality of the calculation:

1) r should be 0.0
2) P_global_CG( = A*U ) - B should be 0.0
*/

/*
	check = PoisConjPassemble_AMR( P_global_CG, U, dof, block_size);
	if(!check) printf( " Problems with PoisConjPassemble_AMR \n");

	for( j = 0; j < dof; ++j )
	{
		printf( "%4d %14.6f  %14.6f %14.6f  %14.6f  %14.6f %14.6f\n",j,alpha,beta,
			*(U+j),*(r+j),*(P_global_CG+j),*(B+j));
	}
*/

/* Set Pressure equal to the solution U */

	for( j = 0; j < ny_f; ++j )
	{
	    for( i = 0; i < nx_f; ++i )
	    {
		bce  = i + nx_f*j;
		ce  = ny_f*i + j;

		*(prs + ce) = *(U + bce);
	    }
	}

	/*for( j = 0; j < ny_f; ++j )
	{
	    for( i = 0; i < nx_f; ++i )
	    {
		ce  = ny_f*i + j;
		printf( "%4d %4d %14.6e\n", i, j, *(prs + ce));
	    }
	}*/

	free(mem_double);

	return 1;
}


int PoisConjGradCoarse( double *X, double *Y, double *prs, SUBGRID subgrid_f, SUBGRID subgrid_c)
{
/* This function does memory allocation and uses the conjugate gradient
   method to solve the linear system which arises from the discrete
   Poisson equation.

                  Last Update 3/28/08

   It is taken from the algorithm 10.3.1 given in "Matrix Computations",
   by Golub, page 534.
*/
	int i, j, k, sofmf, ptr_inc, dof, block_size;
	int check, counter;
	double *mem_double;
	double *p, *P_global_CG, *r, *z;
	double *pre_cond, *B, *U;
	double alpha, alpha2, beta;
	double fdum, fdum2, fdum3;
	int ce, xm1, xm2, xp1, ym1, ym2, yp1, yp2, xm1ym1, xm1yp1, xp1ym1, xp1yp1;
	int bce, bxm1, bxm2, bxm3, bxp1, bxp2, bxp3, bym1, bym2, byp1, byp2,
		bxm1ym1, bxm1yp1, bxp1ym1, bxp1yp1;
	double pxm1, pxp1, pym1, pyp1;
	int  iteration_max, iteration_const, iteration;
	int  nx, ny;
	double tolerance;

	iteration_max = 2000;
	iteration = 0;
	tolerance = 1.0e-13;

	nx = subgrid_c.nx;
	ny = subgrid_c.ny;

	dof = nx*ny;
	block_size = nx;

/* For the doubles */
	sofmf = 7*dof;
	mem_double=(double *)calloc(sofmf,sizeof(double));

	if(!mem_double )
	{
		printf( "failed to allocate memory for doubles\n ");
		exit(1);
	}

/* For the Conjugate Gradient Method doubles */

	                                        ptr_inc = 0;
	p=(mem_double+ptr_inc);                 ptr_inc += dof;
	P_global_CG=(mem_double+ptr_inc);       ptr_inc += dof;
	r=(mem_double+ptr_inc);                 ptr_inc += dof;
	z=(mem_double+ptr_inc);                 ptr_inc += dof;
	B=(mem_double+ptr_inc);                 ptr_inc += dof;
	U=(mem_double+ptr_inc);                 ptr_inc += dof;
	pre_cond=(mem_double+ptr_inc);          ptr_inc += dof;

/* Using Conjugate gradient method to find displacements */

	memset(P_global_CG,0,dof*sof);
	memset(B,0,dof*sof);
	memset(p,0,dof*sof);
	memset(r,0,dof*sof);
	memset(z,0,dof*sof);

/* Create the RHS vector

IMPORTANT!!!

I needed to reorder the indices such that the varaibles which are used
for this calculation, B, U, P_global_CG, p, r, and z follow natural ordering. 
This is why I have both "ce" and "bce" below.  So any time I am dealing
with theses variables, there should be a "b" prefix to the indexing quantity
(e.g.  bxm1, bxm2, byp1, byp2).  Remember this when dealing with these
variables and note especially that "ny*i + j" becomes "i + nx*j".
 */

	for( j = 1; j < ny-1; ++j )
	{
	    for( i = 1; i < nx-1; ++i )
	    {
		   bce  = i + nx*j;
		   ce  = ny*i + j;
		   xm1 = ny*(i-1) + j;
		   xp1 = ny*(i+1) + j;
		   ym1 = ny*i + (j-1);
		   yp1 = ny*i + (j+1);

		   *(B + bce) = -dx/dt*(*(X + ce) - *(X + xm1) + *(Y + ce) - *(Y + ym1));
		   *(U + bce) = *(prs + ce);
	    }
	}

/* The initial setting of "r" is for it to be set to "B" above. */

	/*memcpy( r, B, dof*sizeof(double));*/

/* Use the value of the last pressure as an initial guess. */

	check = PoisConjPassemble( P_global_CG, U, dof, block_size);
	if(!check) printf( " Problems with PoisConjPassemble \n");

	for( i = 0; i < dof; ++i )
	{
		*(r + i) = *(B + i) - *(P_global_CG + i);
	}

	for( k = 0; k < 10; ++k )
	{
	    for( j = 0; j < dof; ++j )
	    {
		*(pre_cond + j) += 1.0 + SMALL;
		*(z + j) = *(r + j)/(*(pre_cond + j));
		*(p+j) = *(z+j);
	    }

	    check = Coarse_Boundary( p, subgrid_c, subgrid_f);
	    if(!check) printf( " Problems with Perimeter_Boundary\n");

	    check = Coarse_Boundary( r, subgrid_c, subgrid_f);
	    if(!check) printf( " Problems with Perimeter_Boundary\n");

	    alpha = 0.0;
	    alpha2 = 0.0;
	    beta = 0.0;
	    fdum2 = 1000.0;
	    counter = 0;
	    check = dotX(&fdum, r, z, dof);

	    printf("\n iteration %3d iteration max %3d \n", iteration, iteration_max);
	    /*for( iteration = 0; iteration < iteration_max; ++iteration )*/

/* The outer "k" loop is for when the newly calculated pressures are extrapolated to the
   boundaries.  I start the calculation over until the code converges.
*/
	    while( fabs(fdum2) > tolerance && counter < iteration_max )
	    {

		/*printf( "\n %3d %16.8e\n",counter, fdum2);*/
		check = PoisConjPassemble( P_global_CG, p, dof, block_size);
		if(!check) printf( " Problems with PoisConjPassemble \n");

		check = Coarse_Boundary( P_global_CG, subgrid_c, subgrid_f);
		if(!check) printf( " Problems with Perimeter_Boundary\n");

		check = dotX(&alpha2, p, P_global_CG, dof);	
		alpha = fdum/(SMALL + alpha2);
		for( j = 0; j < dof; ++j )
		{
		    /*printf( "%4d %14.5e  %14.5e  %14.5e  %14.5e  %14.5e %14.5e\n",j,alpha,
			beta,*(U+j),*(r+j),*(P_global_CG+j),*(p+j));*/
		    *(U+j) += alpha*(*(p+j));
		    *(r+j) -= alpha*(*(P_global_CG+j));
		    *(z + j) = *(r + j)/(*(pre_cond + j));
		}

		check = dotX(&fdum2, r, z, dof);
		beta = fdum2/(SMALL + fdum);
		fdum = fdum2;
		
		for( j = 0; j < dof; ++j )
		{
		    /*printf("\n  %3d %12.7f  %14.5f ",j,*(U+j),*(P_global_CG+j));*/
		    /*printf( "%4d %14.5f  %14.5f  %14.5f  %14.5f %14.5f\n",j,alpha,
			*(U+j),*(r+j),*(P_global_CG+j),*(B+j));
		    printf( "%4d %14.8f  %14.8f  %14.8f  %14.8f %14.8f\n",j,
			*(U+j)*bet,*(r+j)*bet,*(P_global_CG+j)*alp/(*(mass+j)),
			*(B+j)*alp/(*(mass+j)));*/
		    *(p+j) = *(z+j)+beta*(*(p+j));
		}

		check = Coarse_Boundary( p, subgrid_c, subgrid_f);
		if(!check) printf( " Problems with Perimeter_Boundary\n");

		++counter;
	    }

/* Take the newly calculated pressures and extrapoplate them to the boundary.
   This, combined with the previous calculation of pressure, represented by U,
   is used as an initial guess and the calculation begins again.

   Note that the indexing variables "ce" are different from how they are outside
   the loop. 
*/

#if 0
/* This is for the first node at the bottom right. */

	    i = nx; j = 0;
	    bce = i + nx*j;
	    xp1 = (i+1) + nx*j;
	    ym1 = (i+1) + nx*(j-1);
	    *(r + bce) = *(U + xp1) + *(U + yp1);
#endif

/* This is for the nodes at the right exit.
   Note that I have to shift the indices of i and j when indexing for "prs".
*/

	    /* memset(r,0,dof*sof); */

#if 1
	    i = 0;
	    i = nx - 1;
	    for( j = 1; j < ny-1; ++j )
	    {
		bce = i + nx*j;
		bxm1 = (i-1) + nx*j;
		bxm2 = (i-2) + nx*j;
		bxm3 = (i-3) + nx*j;
		bxp1 = (i+1) + nx*j;
		bxp2 = (i+2) + nx*j;
		bxp3 = (i+3) + nx*j;
		xp1  = ny*(i+1) + j;

/* Linearlly extrapolate the pressure at the right. */

		/**(U + bce) = 2.0*(*(U + bxm2)) - *(U + bxm3);*/

		*(U + bce) = 2.0*(*(U + bxm1)) - *(U + bxm2);

		/* *(prs + xp1) = 2.0*(*(U + bxm2)) - *(U + bxm3);*/
		/**(prs + xp1) = 2.0*(*(U + bce)) - *(U + bxm1);*/
		/*printf( "%4d %4d %14.6e\n", i+2, j+1, *(prs + xp1));*/
	    }
#endif

/* This is for the nodes at the top and bottom exits. */

	    for( i = 0; i < nx-1; ++i )
	    {
	        j = ny - 1;
		bce = i + nx*j;
		bym1 = i + nx*(j-1);
		bym2 = i + nx*(j-2);
		ce  = ny*i + j;

/* Linearlly extrapolate the pressure at the top. */

		/**(prs + ce) = 2.0*(*(U + bym1)) - *(U + bym2);
		*(prs + ce) = *(U + bym1);*/
		/*printf( "%4d %4d %14.6e\n", i+1, j+2, *(prs + yp1));*/

		*(U + bce) = *(U + bym1);

	        j = 0;
		bce = i + nx*j;
		byp1 = i + nx*(j+1);
		byp2 = i + nx*(j+2);
		ce  = ny*i + j;

/* Linearlly extrapolate the pressure at the bottom. */

		/**(prs + ce) = 2.0*(*(U + byp1)) - *(U + byp2);
		*(prs + ce) = *(U + byp1);*/
		/*printf( "%4d %4d %14.6e\n", i+1, j, *(prs + ym1));*/

		*(U + bce) = *(U + byp1);

	    }
	    /*if(k == 0) memset(U,0,dof*sof);*/

/* Use the new value of U as an initial guess and then begin the calculation again. */

	    check = PoisConjPassemble( P_global_CG, U, dof, block_size);
	    if(!check) printf( " Problems with PoisConjPassemble \n");

	    fdum3 = 0.0;
	    for( i = 0; i < dof; ++i )
	    {
		*(r + i) = *(B + i) - *(P_global_CG + i);
		*(B + i) = *(P_global_CG + i);
		fdum3 += fabs(*(r + i));
	    }
	    if(fdum3 < 1.0e-8) break;
	    printf( "\n fffff  %16.8e\n", fdum3);
	}


	if(counter > iteration_max - 1 )
	{
		printf( "\nMaximum iterations %4d reached.  Residual is: %16.8e\n",counter,fdum2);
		printf( "Problem may not have converged during Conj. Grad.\n");
	}

/*
The lines below are for testing the quality of the calculation:

1) r should be 0.0
2) P_global_CG( = A*U ) - B should be 0.0
*/

/*
	check = PoisConjPassemble( P_global_CG, U, dof, block_size);
	if(!check) printf( " Problems with PoisConjPassemble \n");

	for( j = 0; j < dof; ++j )
	{
		printf( "%4d %14.6f  %14.6f %14.6f  %14.6f  %14.6f %14.6f\n",j,alpha,beta,
			*(U+j),*(r+j),*(P_global_CG+j),*(B+j));
	}
*/

/* Set Pressure equal to the solution U */

	for( j = 0; j < ny; ++j )
	{
	    for( i = 0; i < nx; ++i )
	    {
		bce  = i + nx*j;
		ce  = ny*i + j;

		*(prs + ce) = *(U + bce);
	    }
	}

	/*for( j = 0; j < ny; ++j )
	{
	    for( i = 0; i < nx; ++i )
	    {
		ce  = ny*i + j;
		printf( "%4d %4d %14.6e\n", i, j, *(prs + ce));
	    }
	}*/

	free(mem_double);

	return 1;
}


