/*
    This library function sets the boundary condition variables for the
    program which performs finite difference analysis for a compressible
    CFD flow field.

    SLFCFD source file
    Version:  1.1
    Copyright (C) 2005, 2006, 2007, 2008  San Le

    The source code contained in this file is released under the
    terms of the GNU Library General Public License.

*/

#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "cfdconst.h"
#include "cfdstruct.h"

#if COMPRESSIBLE
#define NODE    COM_NODE
#endif
#if INCOMPRESSIBLE
#define NODE    INC_NODE
#endif
#if GUI
#define NODE    GR_NODE
#endif

int numnp, sof, nx, ny, nz, num_bodies;
double dx, dy, dz;
int flag_2D, flag_3D;
double tolerance;

int bcset ( BOUND *bc, int *mem_int, NODE *node)
{
/* Note:  The reason bc is passed as *bc is because it is in this subroutine that I do memory
          allocation for the elements of bc which pertain to recording the nodes for a particular
          boundary type.  This creates the unfortunate need for using the notation bc[0].* rather
          than just bc.* (e.g. bc[0].num_none[0] rather than bc.num_none[0]).  The reason I
          wanted to do the memory allocation for bc here rather than the main where I normally do
          it is because I don't know the number of nodes pertaining to each boundary until I do
          the first loop below.  Rather than counting and passing it all back to the main program,
          I have decided to everything for bc here.

                     Updated 2/9/08

*/
	int i, j, k, ptr_inc;
	int dum, node_num, *node_counter, ce;
	int dum0, dum1, dum2, dum3, dum4, dum5, dum6, dum7, dum8, dum9,
	    dum10, dum11, dum12, dum13, dum14, dum15, dum16, dum17, dum18, dum19,
	    dum20, dum21, dum22, dum23, dum24, dum25, dum26, dum27, dum28, dum29,
	    dum30, dum31, dum32, dum33, dum34, dum35, dum36, dum37, dum38, dum39,
	    dum40, dum41, dum42, dum43;

	for( i = 0; i < numnp; ++i)
	{
/* Count the number of nodes for a particular boundary condition type. */

		if( node[i].bc_type == 0 ) bc[0].num_none[0]      += 1; /* None */
		if( node[i].bc_type == 1 ) bc[0].num_symm[0].xm   += 1; /* symmetric in x - */
		if( node[i].bc_type == 2 ) bc[0].num_symm[0].xp   += 1; /* symmetric in x + */
		if( node[i].bc_type == 3 ) bc[0].num_symm[0].ym   += 1; /* symmetric in y - */
		if( node[i].bc_type == 4 ) bc[0].num_symm[0].yp   += 1; /* symmetric in y + */
		if( node[i].bc_type == 5 ) bc[0].num_symm[0].zm   += 1; /* symmetric in z - */
		if( node[i].bc_type == 6 ) bc[0].num_symm[0].zp   += 1; /* symmetric in z + */
		if( node[i].bc_type == 7 ) bc[0].num_far[0].xm    += 1; /* far field in x - */
		if( node[i].bc_type == 8 ) bc[0].num_far[0].xp    += 1; /* far field in x + */
		if( node[i].bc_type == 9 ) bc[0].num_far[0].ym    += 1; /* far field in y - */
		if( node[i].bc_type == 10) bc[0].num_far[0].yp    += 1; /* far field in y + */
		if( node[i].bc_type == 11) bc[0].num_far[0].zm    += 1; /* far field in z - */
		if( node[i].bc_type == 12) bc[0].num_far[0].zp    += 1; /* far field in z + */
		if( node[i].bc_type == 13) bc[0].num_free[0].xm   += 1; /* free in x - */
		if( node[i].bc_type == 14) bc[0].num_free[0].xp   += 1; /* free in x + */
		if( node[i].bc_type == 15) bc[0].num_free[0].ym   += 1; /* free in y - */
		if( node[i].bc_type == 16) bc[0].num_free[0].yp   += 1; /* free in y + */
		if( node[i].bc_type == 17) bc[0].num_free[0].zm   += 1; /* free in z - */
		if( node[i].bc_type == 18) bc[0].num_free[0].zp   += 1; /* free in z + */
		if( node[i].bc_type == 19) bc[0].num_inflow[0].xm += 1; /* inflow in x - */
		if( node[i].bc_type == 20) bc[0].num_inflow[0].xp += 1; /* inflow in x + */
		if( node[i].bc_type == 21) bc[0].num_inflow[0].ym += 1; /* inflow in y - */
		if( node[i].bc_type == 22) bc[0].num_inflow[0].yp += 1; /* inflow in y + */
		if( node[i].bc_type == 23) bc[0].num_inflow[0].zm += 1; /* inflow in z - */
		if( node[i].bc_type == 24) bc[0].num_inflow[0].zp += 1; /* inflow in z + */
		if( node[i].bc_type == 25) bc[0].num_frestr[0].xm += 1; /* freestream in x - */
		if( node[i].bc_type == 26) bc[0].num_frestr[0].xp += 1; /* freestream in x + */
		if( node[i].bc_type == 27) bc[0].num_frestr[0].ym += 1; /* freestream in y - */
		if( node[i].bc_type == 28) bc[0].num_frestr[0].yp += 1; /* freestream in y + */
		if( node[i].bc_type == 29) bc[0].num_frestr[0].zm += 1; /* freestream in z - */
		if( node[i].bc_type == 30) bc[0].num_frestr[0].zp += 1; /* freestream in z + */
		if( node[i].bc_type == 31) bc[0].num_free2[0].xm  += 1; /* free 2 in x - */
		if( node[i].bc_type == 32) bc[0].num_free2[0].xp  += 1; /* free 2 in x + */
		if( node[i].bc_type == 33) bc[0].num_free2[0].ym  += 1; /* free 2 in y - */
		if( node[i].bc_type == 34) bc[0].num_free2[0].yp  += 1; /* free 2 in y + */
		if( node[i].bc_type == 35) bc[0].num_free2[0].zm  += 1; /* free 2 in z - */
		if( node[i].bc_type == 36) bc[0].num_free2[0].zp  += 1; /* free 2 in z + */
		if( node[i].bc_type == 37) bc[0].num_noslp[0].xm  += 1; /* noslip in x - */
		if( node[i].bc_type == 38) bc[0].num_noslp[0].xp  += 1; /* noslip in x + */
		if( node[i].bc_type == 39) bc[0].num_noslp[0].ym  += 1; /* noslip in y - */
		if( node[i].bc_type == 40) bc[0].num_noslp[0].yp  += 1; /* noslip in y + */
		if( node[i].bc_type == 41) bc[0].num_noslp[0].zm  += 1; /* noslip in z - */
		if( node[i].bc_type == 42) bc[0].num_noslp[0].zp  += 1; /* noslip in z + */

		if( node[i].bc_type == 43) /* corner noslip in x -, in y - */
		{
		    bc[0].num_noslp[0].xm  += 1; /* noslip in x - */
		    bc[0].num_noslp[0].ym  += 1; /* noslip in y - */
		}
		if( node[i].bc_type == 44) /* corner noslip in x -, in y + */
		{
		    bc[0].num_noslp[0].xm  += 1; /* noslip in x - */
		    bc[0].num_noslp[0].yp  += 1; /* noslip in y + */
		}
		if( node[i].bc_type == 45) /* corner noslip in x +, in y - */
		{
		    bc[0].num_noslp[0].xp  += 1; /* noslip in x + */
		    bc[0].num_noslp[0].ym  += 1; /* noslip in y - */
		}
		if( node[i].bc_type == 46) /* corner noslip in x +, in y + */
		{
		    bc[0].num_noslp[0].xp  += 1; /* noslip in x + */
		    bc[0].num_noslp[0].yp  += 1; /* noslip in y + */
		}

		if( node[i].bc_type == 63) bc[0].num_prs0[0]      += 1; /* pressure node set to 0.0 */
	}

/*  Allocate the memory for the part of bc which records the nodes for a particular boundary
    condition type.
*/
	                                       ptr_inc = 0;
	bc[0].none=(mem_int+ptr_inc);          ptr_inc += bc[0].num_none[0];       /* None */
	bc[0].symm.xm=(mem_int+ptr_inc);       ptr_inc += bc[0].num_symm[0].xm;    /* symmetric in x - */
	bc[0].symm.xp=(mem_int+ptr_inc);       ptr_inc += bc[0].num_symm[0].xp;    /* symmetric in x + */
	bc[0].symm.ym=(mem_int+ptr_inc);       ptr_inc += bc[0].num_symm[0].ym;    /* symmetric in y - */
	bc[0].symm.yp=(mem_int+ptr_inc);       ptr_inc += bc[0].num_symm[0].yp;    /* symmetric in y + */
	bc[0].symm.zm=(mem_int+ptr_inc);       ptr_inc += bc[0].num_symm[0].zm;    /* symmetric in z - */
	bc[0].symm.zp=(mem_int+ptr_inc);       ptr_inc += bc[0].num_symm[0].zp;    /* symmetric in z + */
	bc[0].far.xm=(mem_int+ptr_inc);        ptr_inc += bc[0].num_far[0].xm;     /* far field in x - */
	bc[0].far.xp=(mem_int+ptr_inc);        ptr_inc += bc[0].num_far[0].xp;     /* far field in x + */
	bc[0].far.ym=(mem_int+ptr_inc);        ptr_inc += bc[0].num_far[0].ym;     /* far field in y - */
	bc[0].far.yp=(mem_int+ptr_inc);        ptr_inc += bc[0].num_far[0].yp;     /* far field in y + */
	bc[0].far.zm=(mem_int+ptr_inc);        ptr_inc += bc[0].num_far[0].zm;     /* far field in z - */
	bc[0].far.zp=(mem_int+ptr_inc);        ptr_inc += bc[0].num_far[0].zp;     /* far field in z + */
	bc[0].free.xm=(mem_int+ptr_inc);       ptr_inc += bc[0].num_free[0].xm;    /* free in x - */
	bc[0].free.xp=(mem_int+ptr_inc);       ptr_inc += bc[0].num_free[0].xp;    /* free in x + */
	bc[0].free.ym=(mem_int+ptr_inc);       ptr_inc += bc[0].num_free[0].ym;    /* free in y - */
	bc[0].free.yp=(mem_int+ptr_inc);       ptr_inc += bc[0].num_free[0].yp;    /* free in y + */
	bc[0].free.zm=(mem_int+ptr_inc);       ptr_inc += bc[0].num_free[0].zm;    /* free in z - */
	bc[0].free.zp=(mem_int+ptr_inc);       ptr_inc += bc[0].num_free[0].zp;    /* free in z + */
	bc[0].inflow.xm=(mem_int+ptr_inc);     ptr_inc += bc[0].num_inflow[0].xm;  /* inflow in x - */
	bc[0].inflow.xp=(mem_int+ptr_inc);     ptr_inc += bc[0].num_inflow[0].xp;  /* inflow in x + */
	bc[0].inflow.ym=(mem_int+ptr_inc);     ptr_inc += bc[0].num_inflow[0].ym;  /* inflow in y - */
	bc[0].inflow.yp=(mem_int+ptr_inc);     ptr_inc += bc[0].num_inflow[0].yp;  /* inflow in y + */
	bc[0].inflow.zm=(mem_int+ptr_inc);     ptr_inc += bc[0].num_inflow[0].zm;  /* inflow in z - */
	bc[0].inflow.zp=(mem_int+ptr_inc);     ptr_inc += bc[0].num_inflow[0].zp;  /* inflow in z + */

	bc[0].frestr.xm=(mem_int+ptr_inc);     ptr_inc += bc[0].num_frestr[0].xm;  /* freestream in x - */
	bc[0].frestr.xp=(mem_int+ptr_inc);     ptr_inc += bc[0].num_frestr[0].xp;  /* freestream in x + */
	bc[0].frestr.ym=(mem_int+ptr_inc);     ptr_inc += bc[0].num_frestr[0].ym;  /* freestream in y - */
	bc[0].frestr.yp=(mem_int+ptr_inc);     ptr_inc += bc[0].num_frestr[0].yp;  /* freestream in y + */
	bc[0].frestr.zm=(mem_int+ptr_inc);     ptr_inc += bc[0].num_frestr[0].zm;  /* freestream in z - */
	bc[0].frestr.zp=(mem_int+ptr_inc);     ptr_inc += bc[0].num_frestr[0].zp;  /* freestream in z + */
	bc[0].free2.xm=(mem_int+ptr_inc);      ptr_inc += bc[0].num_free2[0].xm;   /* free 2 in x - */
	bc[0].free2.xp=(mem_int+ptr_inc);      ptr_inc += bc[0].num_free2[0].xp;   /* free 2 in x + */
	bc[0].free2.ym=(mem_int+ptr_inc);      ptr_inc += bc[0].num_free2[0].ym;   /* free 2 in y - */
	bc[0].free2.yp=(mem_int+ptr_inc);      ptr_inc += bc[0].num_free2[0].yp;   /* free 2 in y + */
	bc[0].free2.zm=(mem_int+ptr_inc);      ptr_inc += bc[0].num_free2[0].zm;   /* free 2 in z - */
	bc[0].free2.zp=(mem_int+ptr_inc);      ptr_inc += bc[0].num_free2[0].zp;   /* free 2 in z + */
	bc[0].noslp.xm=(mem_int+ptr_inc);      ptr_inc += bc[0].num_noslp[0].xm;   /* noslip in x - */
	bc[0].noslp.xp=(mem_int+ptr_inc);      ptr_inc += bc[0].num_noslp[0].xp;   /* noslip in x + */
	bc[0].noslp.ym=(mem_int+ptr_inc);      ptr_inc += bc[0].num_noslp[0].ym;   /* noslip in y - */
	bc[0].noslp.yp=(mem_int+ptr_inc);      ptr_inc += bc[0].num_noslp[0].yp;   /* noslip in y + */
	bc[0].noslp.zm=(mem_int+ptr_inc);      ptr_inc += bc[0].num_noslp[0].zm;   /* noslip in z - */
	bc[0].noslp.zp=(mem_int+ptr_inc);      ptr_inc += bc[0].num_noslp[0].zp;   /* noslip in z + */

	bc[0].prs0=(mem_int+ptr_inc);          ptr_inc += bc[0].num_prs0[0];       /* pressure node set to 0.0 */

	dum0  = 0, dum1  = 0, dum2  = 0, dum3  = 0, dum4  = 0;
	dum5  = 0, dum6  = 0, dum7  = 0, dum8  = 0, dum9  = 0;
	dum10 = 0, dum11 = 0, dum12 = 0, dum13 = 0, dum14 = 0;
	dum15 = 0, dum16 = 0, dum17 = 0, dum18 = 0, dum19 = 0;
	dum20 = 0, dum21 = 0, dum22 = 0, dum23 = 0, dum24 = 0;
	dum25 = 0, dum26 = 0, dum27 = 0, dum28 = 0, dum29 = 0;
	dum30 = 0, dum31 = 0, dum32 = 0, dum33 = 0, dum34 = 0;
	dum35 = 0, dum36 = 0, dum37 = 0, dum38 = 0, dum39 = 0;
	dum40 = 0, dum41 = 0, dum42 = 0, dum43 = 0;

	for( i = 0; i < numnp; ++i)
	{
/*  Record the nodes for a particular boundary condition type.
*/
		if( node[i].bc_type == 0 ) {
			bc[0].none[dum0]      = i;
			++dum0;
		}
		if( node[i].bc_type == 1 ) {
			bc[0].symm.xm[dum1]    = i; /* symmetric in x - */
			++dum1;
		}
		if( node[i].bc_type == 2 ) {
			bc[0].symm.xp[dum2]    = i; /* symmetric in x + */
			++dum2;
		}
		if( node[i].bc_type == 3 ) {
			bc[0].symm.ym[dum3]    = i; /* symmetric in y - */
			++dum3;
		}
		if( node[i].bc_type == 4 ) {
			bc[0].symm.yp[dum4]    = i; /* symmetric in y + */
			++dum4;
		}
		if( node[i].bc_type == 5 ) {
			bc[0].symm.zm[dum5]    = i; /* symmetric in z - */
			++dum5;
		}
		if( node[i].bc_type == 6 ) {
			bc[0].symm.zp[dum6]    = i; /* symmetric in z + */
			++dum6;
		}
		if( node[i].bc_type == 7 ) {
			bc[0].far.xm[dum7]     = i; /* far field in x - */
			++dum7;
		}
		if( node[i].bc_type == 8 ) {
			bc[0].far.xp[dum8]     = i; /* far field in x + */
			++dum8;
		}
		if( node[i].bc_type == 9 ) {
			bc[0].far.ym[dum9]     = i; /* far field in y - */
			++dum9;
		}
		if( node[i].bc_type == 10 ) {
			bc[0].far.yp[dum10]    = i; /* far field in y + */
			++dum10;
		}
		if( node[i].bc_type == 11 ) {
			bc[0].far.zm[dum11]    = i; /* far field in z - */
			++dum11;
		}
		if( node[i].bc_type == 12 ) {
			bc[0].far.zp[dum12]    = i; /* far field in z + */
			++dum12;
		}
		if( node[i].bc_type == 13 ) {
			bc[0].free.xm[dum13]   = i; /* free in x - */
			++dum13;
		}
		if( node[i].bc_type == 14 ) {
			bc[0].free.xp[dum14]   = i; /* free in x + */
			++dum14;
		}
		if( node[i].bc_type == 15 ) {
			bc[0].free.ym[dum15]   = i; /* free in y - */
			++dum15;
		}
		if( node[i].bc_type == 16 ) {
			bc[0].free.yp[dum16]   = i; /* free in y + */
			++dum16;
		}
		if( node[i].bc_type == 17 ) {
			bc[0].free.zm[dum17]   = i; /* free in z - */
			++dum17;
		}
		if( node[i].bc_type == 18 ) {
			bc[0].free.zp[dum18]   = i; /* free in z + */
			++dum18;
		}
		if( node[i].bc_type == 19) {
			bc[0].inflow.xm[dum19] = i; /* inflow in x - */
			++dum19;
		}
		if( node[i].bc_type == 20) {
			bc[0].inflow.xp[dum20] = i; /* inflow in x + */
			++dum20;
		}
		if( node[i].bc_type == 21) {
			bc[0].inflow.ym[dum21] = i; /* inflow in y - */
			++dum21;
		}
		if( node[i].bc_type == 22) {
			bc[0].inflow.yp[dum22] = i; /* inflow in y + */
			++dum22;
		}
		if( node[i].bc_type == 23) {
			bc[0].inflow.zm[dum23] = i; /* inflow in z - */
			++dum23;
		}
		if( node[i].bc_type == 24) {
			bc[0].inflow.zp[dum24] = i; /* inflow in z + */
			++dum24;
		}
		if( node[i].bc_type == 25) {
			bc[0].frestr.xm[dum25] = i; /* freestream in x - */
			++dum25;
		}
		if( node[i].bc_type == 26) {
			bc[0].frestr.xp[dum26] = i; /* freestream in x + */
			++dum26;
		}
		if( node[i].bc_type == 27) {
			bc[0].frestr.ym[dum27] = i; /* freestream in y - */
			++dum27;
		}
		if( node[i].bc_type == 28) {
			bc[0].frestr.yp[dum28] = i; /* freestream in y + */
			++dum28;
		}
		if( node[i].bc_type == 29) {
			bc[0].frestr.zm[dum29] = i; /* freestream in z - */
			++dum29;
		}
		if( node[i].bc_type == 30) {
			bc[0].frestr.zp[dum30] = i; /* freestream in z + */
			++dum30;
		}
		if( node[i].bc_type == 31) {
			bc[0].free2.xm[dum31] = i; /* free 2 in x - */
			++dum31;
		}
		if( node[i].bc_type == 32) {
			bc[0].free2.xp[dum32] = i; /* free 2 in x + */
			++dum32;
		}
		if( node[i].bc_type == 33) {
			bc[0].free2.ym[dum33] = i; /* free 2 in y - */
			++dum33;
		}
		if( node[i].bc_type == 34) {
			bc[0].free2.yp[dum34] = i; /* free 2 in y + */
			++dum34;
		}
		if( node[i].bc_type == 35) {
			bc[0].free2.zm[dum35] = i; /* free 2 in z - */
			++dum35;
		}
		if( node[i].bc_type == 36) {
			bc[0].free2.zp[dum36] = i; /* free 2 in z + */
			++dum36;
		}
		if( node[i].bc_type == 37) {
			bc[0].noslp.xm[dum37] = i; /* noslip in x - */
			++dum37;
		}
		if( node[i].bc_type == 38) {
			bc[0].noslp.xp[dum38] = i; /* noslip in x + */
			++dum38;
		}
		if( node[i].bc_type == 39) {
			bc[0].noslp.ym[dum39] = i; /* noslip in y - */
			++dum39;
		}
		if( node[i].bc_type == 40) {
			bc[0].noslp.yp[dum40] = i; /* noslip in y + */
			++dum40;
		}
		if( node[i].bc_type == 41) {
			bc[0].noslp.zm[dum41] = i; /* noslip in z - */
			++dum41;
		}
		if( node[i].bc_type == 42) {
			bc[0].noslp.zp[dum42] = i; /* noslip in z + */
			++dum42;
		}

/* For the corner nodes */

		if( node[i].bc_type == 43) {
			bc[0].noslp.xm[dum37] = i; /* noslip in x - */
			++dum37;
			bc[0].noslp.ym[dum39] = i; /* noslip in y - */
			++dum39;
		}
		if( node[i].bc_type == 44) {
			bc[0].noslp.xm[dum37] = i; /* noslip in x - */
			++dum37;
			bc[0].noslp.yp[dum40] = i; /* noslip in y + */
			++dum40;
		}
		if( node[i].bc_type == 45) {
			bc[0].noslp.xp[dum38] = i; /* noslip in x + */
			++dum38;
			bc[0].noslp.ym[dum39] = i; /* noslip in y - */
			++dum39;
		}
		if( node[i].bc_type == 46) {
			bc[0].noslp.xp[dum38] = i; /* noslip in x + */
			++dum38;
			bc[0].noslp.yp[dum40] = i; /* noslip in y + */
			++dum40;
		}

		if( node[i].bc_type == 63) {
			bc[0].prs0[dum43] = i; /* pressure node set to 0.0 */
			++dum43;
		}
	}

	return 1;
}


int bcset_noslip ( BOUND *bc, BODY *body, int *mem_int, NODE *node)
{
/* These are the additional boundary conditions related to the noslip BC.  This applies
   to finding the noslip pressure corner nodes as well as nodes internal to a simply
   connected body.

                     Updated 2/9/08

*/
	int i, j, k, i2, j2, k2, i3, j3, k3, ptr_inc;
	int dum, bod_num, node_num, *node_counter, ce;
	int dum0, dum1, dum2, dum3, dum4, dum5, dum6, dum7, dum8;

	node_counter=(int *)calloc(numnp,sizeof(int));
	memset(node_counter,0,numnp*sizeof(int));

/* Find the nodes where noslip nodes in x overlap with noslip nodes in y.  This is needed
   because the calculation of pressure on noslip regions involve the use of velocity.
   Where there is overlap, I have to use velocities in both directions.
*/

	ptr_inc = 0;
	bc[0].noslp_prs.xpyp=(mem_int+ptr_inc);

	for( dum = 0; dum < bc[0].num_noslp[0].xp; ++dum)
	{
	   node_num = bc[0].noslp.xp[dum];
	   i = node[node_num].i;
	   j = node[node_num].j;
	   ce   =  ny*i + j;

	   node_counter[ce] += 1;
	}

	dum0 = 0;
	bc[0].num_noslp_prs[0].xpyp = 0;
	for( dum = 0; dum < bc[0].num_noslp[0].yp; ++dum)
	{
	   node_num = bc[0].noslp.yp[dum];
	   i = node[node_num].i;
	   j = node[node_num].j;
	   ce   =  ny*i + j;

	   node_counter[ce] += 1;
	   if(node_counter[ce] > 1)
	   {
		bc[0].num_noslp_prs[0].xpyp += 1;
		bc[0].noslp_prs.xpyp[dum0] = ce; /* pressure corner node set to 0.0 */
		++dum0;
	   }
	   node_counter[ce] -= 1;
	}

	ptr_inc += bc[0].num_noslp_prs[0].xpyp;
	bc[0].noslp_prs.xpym=(mem_int+ptr_inc);

	dum0 = 0;
	bc[0].num_noslp_prs[0].xpym = 0;
	for( dum = 0; dum < bc[0].num_noslp[0].ym; ++dum)
	{
	   node_num = bc[0].noslp.ym[dum];
	   i = node[node_num].i;
	   j = node[node_num].j+1; /* Note I set this at 1 node more than the last node. */
	   ce   =  ny*i + j;

	   node_counter[ce] += 1;
	   if(node_counter[ce] > 1)
	   {
		bc[0].num_noslp_prs[0].xpym += 1;
		bc[0].noslp_prs.xpym[dum0] = ce; /* pressure corner node set to 0.0 */
		++dum0;
	   }
	   node_counter[ce] -= 1;
	}

	ptr_inc += bc[0].num_noslp_prs[0].xpym;
	bc[0].noslp_prs.xmyp=(mem_int+ptr_inc);

	memset(node_counter,0,numnp*sizeof(int));

	for( dum = 0; dum < bc[0].num_noslp[0].xm; ++dum)
	{
	   node_num = bc[0].noslp.xm[dum];
	   i = node[node_num].i+1; /* Note I set this at 1 node more than the last node. */
	   j = node[node_num].j;
	   ce   =  ny*i + j;

	   node_counter[ce] += 1;
	}

	dum0 = 0;
	bc[0].num_noslp_prs[0].xmyp = 0;
	for( dum = 0; dum < bc[0].num_noslp[0].yp; ++dum)
	{
	   node_num = bc[0].noslp.yp[dum];
	   i = node[node_num].i;
	   j = node[node_num].j;
	   ce   =  ny*i + j;

	   node_counter[ce] += 1;
	   if(node_counter[ce] > 1)
	   {
		bc[0].num_noslp_prs[0].xmyp += 1;
		bc[0].noslp_prs.xmyp[dum0] = ce; /* pressure corner node set to 0.0 */
		++dum0;
	   }
	   node_counter[ce] -= 1;
	}

	ptr_inc += bc[0].num_noslp_prs[0].xmyp;
	bc[0].noslp_prs.xmym=(mem_int+ptr_inc);

	dum0 = 0;
	bc[0].num_noslp_prs[0].xmym = 0;
	for( dum = 0; dum < bc[0].num_noslp[0].ym; ++dum)
	{
	   node_num = bc[0].noslp.ym[dum];
	   i = node[node_num].i;
	   j = node[node_num].j+1; /* Note I set this at 1 node more than the last node. */
	   ce   =  ny*i + j;

	   node_counter[ce] += 1;
	   if(node_counter[ce] > 1)
	   {
		bc[0].num_noslp_prs[0].xmym += 1;
		bc[0].noslp_prs.xmym[dum0] = ce; /* pressure corner node set to 0.0 */
		++dum0;
	   }
	   node_counter[ce] -= 1;
	}

/* Find the nodes internal to noslip nodes which define the outer boundary of a
   body.  These nodes are needed since velocities internal to a body are set
   to zero.  Currently, I don't think that this is the optimal way to find
   the internal nodes since for a heavily countoured body, the algorithm
   would not catch all the curves that are away from the ends.  For instance
   a body which looks like:
                   _______
        ___       /      |
       /   \_____/       |
      |                  |
      |__________________|

   the way I calculate the vel0.x nodes, based on the algorithm, would likely
   look like:
                   _______
        ___       /++++++|
       /+++\+++++/+++++++|
      |++++++++++++++++++|
      |++++++++++++++++++|

   where the "+" represents the calculated vel0.x nodes.  As you can see, the
   dip in the center of the body is missed.  The same thing happens in the
   vertical direction for vel0.y.  I think I can resolve this problem by finding
   where the vel0.x and vel0.y nodes overlap, and calling those nodes something
   like vel0.xy nodes(This naming may help if I expand to 3D).  Finding the outline
   of a body for display may also suffer in the same way from this issue of too
   many contours.

   The code below is somewhat complex in terms of the logistics of determing
   which nodes are associated with which body.  This is why every noslip node
   also has an associate number indicating which body it belongs to, and this
   is recorded with node[node_num].nsid.
*/


	ptr_inc = numnp;
	bc[0].vel0.x=(mem_int+ptr_inc);

	bc[0].num_vel0[0].x = 0;
	dum3 = 0;
	for( bod_num = 1; bod_num <= num_bodies; ++bod_num)
	{
	    dum4 = 0;
	    for( dum2 = 0; dum2 < bc[0].num_noslp[0].xm; ++dum2)
	    {
		node_num = bc[0].noslp.xm[dum2];
		i = node[node_num].i;
		j = node[node_num].j;
		ce = ny*i + j;

/* The body structure is used to draw the outline of a body.*/ 

		body[bod_num-1].nd[dum4].j = j;
		body[bod_num-1].nd[dum4].i1 = i;
		body[bod_num-1].nd[dum4].i2 = i;
		body[bod_num-1].nd[dum4].index = dum4;

		if( node[node_num].nsid == bod_num)
		{
		    for( dum = 0; dum < bc[0].num_noslp[0].xp; ++dum)
		    {
			node_num = bc[0].noslp.xp[dum];
			i2 = node[node_num].i;
			j2 = node[node_num].j;
			ce   =  ny*i2 + j2;

			if( node[node_num].nsid == bod_num)
			{
			    if( j == j2)
			    {
				for( i3 = i; i3 <= i2; ++i3)
				{
				    ce = ny*i3 + j2;
				    bc[0].vel0.x[dum3] = ce;
				    ++dum3;
				}
				body[bod_num-1].nd[dum4].i2 = i2;
			    }
			}
		    }
		    ++dum4;
		}
	    }
	    body[bod_num-1].num_points = dum4;
	}
	bc[0].num_vel0[0].x = dum3;

	ptr_inc += bc[0].num_vel0[0].x;
	bc[0].vel0.y=(mem_int+ptr_inc);

	bc[0].num_vel0[0].y = 0;
	dum3 = 0;
	for( bod_num = 1; bod_num <= num_bodies; ++bod_num)
	{
	    for( dum2 = 0; dum2 < bc[0].num_noslp[0].ym; ++dum2)
	    {
		node_num = bc[0].noslp.ym[dum2];
		i = node[node_num].i;
		j = node[node_num].j;
		ce = ny*i + j;

		if( node[node_num].nsid == bod_num)
		{
		    for( dum = 0; dum < bc[0].num_noslp[0].yp; ++dum)
		    {
			node_num = bc[0].noslp.yp[dum];
			i2 = node[node_num].i;
			j2 = node[node_num].j;
			ce =  ny*i2 + j2;

			if( node[node_num].nsid == bod_num)
			{
			    if( i == i2)
			    {
				for( j3 = j; j3 <= j2; ++j3)
				{
				    ce = ny*i2 + j3;
				    bc[0].vel0.y[dum3] = ce;
				    ++dum3;
				}
			    }
			}
		    }
		}
	    }
	}
	bc[0].num_vel0[0].y = dum3;

	free(node_counter);

	return 1;
}
