/* This program reapplies the boundary conditions for the program
   which performs finite difference analysis for a compressible
   CFD flow field.  

    SLFCFD source file
    Version:  1.1
    Copyright (C) 2005, 2006, 2007, 2008  San Le

    The source code contained in this file is released under the
    terms of the GNU Library General Public License.

*/

#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "cfdconst.h"
#include "cfdstruct.h"

extern dof, numel, numnp, sof, nx, ny, nz;
extern int flag_2D, flag_3D;
extern  double dx, dy, dz;

#if 0
int applyBC_inflow (void)
{

/*
   Set inflow boundary conditions.  This function is not needed.  I examined
   Imane's code and removed the application of this boundary condition and the results
   were the same.  Here is the comment I put there:

        Like the boundary condition for the farfield, this boundary condition for the inflow
        is unnecessary.  The value at the nodes at this boundary don't change from what they are
        originally set to, and as you can see from the predictor/corrector steps, the loop does not
        include the nodes located at this boundary. 

             Updated 12/19/05
*/

	return 1;

}

int applyBC_symmetric_internal (void)
{

/* Set symmetric boundary conditions for internal nodes.  These are specified as internal
   nodes because they are part of the predictor corrector steps.  This function is not needed
   because of the application of the other symmetric boundary condition.  Imane says this 
   in her code.  The only boundary condition here would have been that the velocity corresponding
   to the symmetric face would be set to zero.  Essentially, this means that the normal
   component of velocity on the symmetric surface is zero.

             Updated 12/20/05
*/

	return 1;

}

int applyBC_far ( BOUND bc, MATL *matl, COM_NODE *node)
{

/*
   Set farfield boundary conditions.  This function is not needed.  I examined
   Imane's code and removed the application of this boundary condition and the results
   were the same.  Here is the comment I put there:

       This boundary condition is unnecessary.  The value at the nodes at this
       boundary don't change from what they are originally set to, and as you can see from
       the predictor/corrector steps, the loop does not include the nodes located
       at this boundary.

             Updated 12/19/05
*/

	return 1;
}

#endif


int applyBC (BOUND bc, MATL *matl, COM_NODE *node)
{
/* The only 2 boundary conditions which change over time and therefore
   require calculation of the applicable nodes for every time step are
   the free and symmetric boundary conditions.  For these 2 boundary
   condition types, the form which it takes are that the boundary node
   are set to some combination of one or two other nodes.  So within this
   subroutine, all that is calculated are the node numbers, and then
   functions which do the actual calculation are called.  There is a fear
   by me that making these calculations a separate function will be
   numerically expensive but I am keeping it like this because it is
   much clearer.

             Updated 1/4/06
*/ 

	int i, j, k, dum, node_num;
	int ijk, ijk1, ijk2, check;
	double dns, temp, prs, vx, vy, vz, enthal, ene;
	double dnsP, tempP, prsP, vxP, vyP, vzP, enthalP, eneP;
	double dnsC, tempC, prsC, vxC, vyC, vzC, enthalC, eneC;
	double fsign[3];

/* The symmetric boundary conditions are set by setting various quantities of the current
   node to the value of the node which is located 2 nodes in the opposite direction of the
   type of boundary condition being set.  For example, for symmetric nodes in the xm side,
   density would be:

     density(i,j,k) = density(i+2,j,k) 

   This is true for all other quantities except for the corresponding velocity which is
   set to the negative of the node which is located 2 nodes in the opposite direction:

     velocity x(i,j,k) = -velocity x(i+2,j,k) 

   As another example, for symmetric nodes in the yp side, density and velocity would be:

     density(i,j,k) = density(i,j-2,k) 
     velocity y(i,j,k) = -velocity y(i,j-2,k) 

   To make the code shorter, I only do the calculation of the applicable
   node numbers within this function, and have a separate function:

     applyBC_symmetric (MATL *matl, COM_NODE *node, COM_NODE *node2, fsign)

   below do the calculation because it has the same form for all symmetric
   boundary conditions.

*/

/* symmetric in x - */

	for( dum = 0; dum < bc.num_symm[0].xm; ++dum)
	{
	   node_num = bc.symm.xm[dum];
	   i = node[node_num].i;
	   j = node[node_num].j;
	   k = node[node_num].k;
	   ijk2 = nz*ny*(i + 2) + nz*j + k;

	   fsign[0] = -1.0; fsign[1] = 1.0; fsign[2] = 1.0;

	   check = applyBC_symmetric (matl, (node + node_num), (node+ijk2), fsign);
	   if(!check) printf( " Problems with applyBC_symmetric \n");
	}

/* symmetric in x + */

	for( dum = 0; dum < bc.num_symm[0].xp; ++dum)
	{
	   node_num = bc.symm.xp[dum];
	   i = node[node_num].i;
	   j = node[node_num].j;
	   k = node[node_num].k;
	   ijk2 = nz*ny*(i - 2) + nz*j + k;

	   fsign[0] = -1.0; fsign[1] = 1.0; fsign[2] = 1.0;

	   check = applyBC_symmetric (matl, (node + node_num), (node+ijk2), fsign);
	   if(!check) printf( " Problems with applyBC_symmetric \n");
	}

/* symmetric in y - */

	for( dum = 0; dum < bc.num_symm[0].ym; ++dum)
	{
	   node_num = bc.symm.ym[dum];
	   i = node[node_num].i;
	   j = node[node_num].j;
	   k = node[node_num].k;
	   ijk2 = nz*ny*i + nz*(j + 2) + k;

	   fsign[0] = 1.0; fsign[1] = -1.0; fsign[2] = 1.0;

	   check = applyBC_symmetric (matl, (node + node_num), (node+ijk2), fsign);
	   if(!check) printf( " Problems with applyBC_symmetric \n");
	}

/* symmetric in y + */

	for( dum = 0; dum < bc.num_symm[0].yp; ++dum)
	{
	   node_num = bc.symm.yp[dum];
	   i = node[node_num].i;
	   j = node[node_num].j;
	   k = node[node_num].k;
	   ijk2 = nz*ny*i + nz*(j - 2) + k;

	   fsign[0] = 1.0; fsign[1] = -1.0; fsign[2] = 1.0;

	   check = applyBC_symmetric (matl, (node + node_num), (node+ijk2), fsign);
	   if(!check) printf( " Problems with applyBC_symmetric \n");
	}

/* symmetric in z - */

	for( dum = 0; dum < bc.num_symm[0].zm; ++dum)
	{
	   node_num = bc.symm.zm[dum];
	   i = node[node_num].i;
	   j = node[node_num].j;
	   k = node[node_num].k;
	   ijk2 = nz*ny*i + nz*j + (k + 2);

	   fsign[0] = 1.0; fsign[1] = 1.0; fsign[2] = -1.0;

	   check = applyBC_symmetric (matl, (node + node_num), (node+ijk2), fsign);
	   if(!check) printf( " Problems with applyBC_symmetric \n");
	}

/* symmetric in z + */

	for( dum = 0; dum < bc.num_symm[0].zp; ++dum)
	{
	   node_num = bc.symm.zp[dum];
	   i = node[node_num].i;
	   j = node[node_num].j;
	   k = node[node_num].k;
	   ijk2 = nz*ny*i + nz*j + (k - 2);

	   fsign[0] = 1.0; fsign[1] = 1.0; fsign[2] = -1.0;

	   check = applyBC_symmetric (matl, (node + node_num), (node+ijk2), fsign);
	   if(!check) printf( " Problems with applyBC_symmetric \n");
	}


/* The free boundary conditions are set by extrapolating various quantities of the current
   node based on the 2 nodes located in the opposite direction of the type of boundary
   condition being set.  For example, for free nodes in the xm side, density would be:

     density(i,j,k) = 2*density(i+1,j,k) - density(i+2,j,k) 

   This is true for all other quantities except for the corresponding velocity which is
   set to the negative of the node which is located 2 nodes in the opposite direction:

   As another example, for free nodes in the yp side, density would be:

     density(i,j,k) = 2*density(i,j+1,k) - density(i,j+2,k) 

   To make the code shorter, I only do the calculation of the applicable
   node numbers within this function, and have a separate function:

     applyBC_free (COM_NODE *node, COM_NODE *node1, COM_NODE *node2)

   below do the calculation because it has the same form for all free
   boundary conditions.

*/

/* free in x - */

	for( dum = 0; dum < bc.num_free[0].xm; ++dum)
	{
	   node_num = bc.free.xm[dum];
	   i = node[node_num].i;
	   j = node[node_num].j;
	   k = node[node_num].k;
	   ijk1 = nz*ny*(i + 1) + nz*j + k;
	   ijk2 = nz*ny*(i + 2) + nz*j + k;

	   check = applyBC_free ((node + node_num), (node+ijk1), (node+ijk2));
	   if(!check) printf( " Problems with applyBC_free \n");

	}

/* free in x + */

	for( dum = 0; dum < bc.num_free[0].xp; ++dum)
	{
	   node_num = bc.free.xp[dum];
	   i = node[node_num].i;
	   j = node[node_num].j;
	   k = node[node_num].k;
	   ijk1 = nz*ny*(i - 1) + nz*j + k;
	   ijk2 = nz*ny*(i - 2) + nz*j + k;

	   check = applyBC_free ((node + node_num), (node+ijk1), (node+ijk2));
	   if(!check) printf( " Problems with applyBC_free \n");
	}

/* free in y + */

	for( dum = 0; dum < bc.num_free[0].ym; ++dum)
	{
	   node_num = bc.free.ym[dum];
	   i = node[node_num].i;
	   j = node[node_num].j;
	   k = node[node_num].k;
	   ijk1 = nz*ny*i + nz*(j + 1) + k;
	   ijk2 = nz*ny*i + nz*(j + 2) + k;

	   check = applyBC_free ((node + node_num), (node+ijk1), (node+ijk2));
	   if(!check) printf( " Problems with applyBC_free \n");
	}

/* free in y - */

	for( dum = 0; dum < bc.num_free[0].yp; ++dum)
	{
	   node_num = bc.free.yp[dum];
	   i = node[node_num].i;
	   j = node[node_num].j;
	   k = node[node_num].k;
	   ijk1 = nz*ny*i + nz*(j - 1) + k;
	   ijk2 = nz*ny*i + nz*(j - 2) + k;

	   check = applyBC_free ((node + node_num), (node+ijk1), (node+ijk2));
	   if(!check) printf( " Problems with applyBC_free \n");
	}

/* free in z + */

	for( dum = 0; dum < bc.num_free[0].zm; ++dum)
	{
	   node_num = bc.free.zm[dum];
	   i = node[node_num].i;
	   j = node[node_num].j;
	   k = node[node_num].k;
	   ijk1 = nz*ny*i + nz*j + (k + 1);
	   ijk2 = nz*ny*i + nz*j + (k + 2);

	   check = applyBC_free ((node + node_num), (node+ijk1), (node+ijk2));
	   if(!check) printf( " Problems with applyBC_free \n");
	}

/* free in z - */

	for( dum = 0; dum < bc.num_free[0].zp; ++dum)
	{
	   node_num = bc.free.zp[dum];
	   i = node[node_num].i;
	   j = node[node_num].j;
	   k = node[node_num].k;
	   ijk1 = nz*ny*i + nz*j + (k - 1);
	   ijk2 = nz*ny*i + nz*j + (k - 2);

	   check = applyBC_free ((node + node_num), (node+ijk1), (node+ijk2));
	   if(!check) printf( " Problems with applyBC_free \n");
	}


	return 1;
}

int applyBC_symmetric (MATL *matl, COM_NODE *node, COM_NODE *node2, double *fsign)
{
/* The first node above is the one lying on the boundary and "node2" is
   the node located 2 nodes in the opposite direction.  I made

     applyBC_symmetric (MATL *matl, COM_NODE *node, COM_NODE *node2, fsign)

   a separate function because it makes everything more readable and I
   am hoping it won't be too inefficient compared to leaving these
   calculations within the loop where they are originally called.

             Updated 2/2/06
*/

	int i, j, k, dum, node_num;
	double dns, temp, prs, vx, vy, vz, enthal, ene;
	double dnsP, tempP, prsP, vxP, vyP, vzP, enthalP, eneP;
	double dnsC, tempC, prsC, vxC, vyC, vzC, enthalC, eneC;
	double gam, R;

	gam = matl[0].gam;
	R = matl[0].R;

/* For the Current Time Step quantities */

	node[0].dns    = dns    =  node2[0].dns;
	node[0].temp   = temp   =  node2[0].temp;
	node[0].prs    = prs    =  node2[0].prs;
	node[0].vx     = vx     =  fsign[0]*node2[0].vx;
	node[0].vy     = vy     =  fsign[1]*node2[0].vy;
	node[0].vz     = vz     =  fsign[2]*node2[0].vz;
	if(flag_2D) vz = 0.0000;
	node[0].ene    = ene    = R*temp/(gam - 1.0) + 0.5*(vx*vx + vy*vy + vz*vz);
	node[0].enthal = enthal = R*gam*temp/(gam - 1.0) + 0.5*(vx*vx + vy*vy + vz*vz);

/* For the Predictor quantities */

	node[0].dnsP    = dnsP    =  node2[0].dnsP;
	node[0].tempP   = tempP   =  node2[0].tempP;
	node[0].prsP    = prsP    =  node2[0].prsP;
	node[0].vxP     = vxP     =  fsign[0]*node2[0].vxP;
	node[0].vyP     = vyP     =  fsign[1]*node2[0].vyP;
	node[0].vzP     = vzP     =  fsign[2]*node2[0].vzP;
	if(flag_2D) vzP = 0.0000;
	node[0].eneP    = eneP    = R*tempP/(gam - 1.0) + 0.5*(vxP*vxP + vyP*vyP + vzP*vzP);
	node[0].enthalP = enthalP = R*gam*tempP/(gam - 1.0) + 0.5*(vxP*vxP + vyP*vyP + vzP*vzP);

/* For the Corrector quantities */

	node[0].dnsC    = dnsC    =  node2[0].dnsC;
	node[0].tempC   = tempC   =  node2[0].tempC;
	node[0].prsC    = prsC    =  node2[0].prsC;
	node[0].vxC     = vxC     =  fsign[0]*node2[0].vxC;
	node[0].vyC     = vyC     =  fsign[1]*node2[0].vyC;
	node[0].vzC     = vzC     =  fsign[2]*node2[0].vzC;
	if(flag_2D) vzC = 0.0000;
	node[0].eneC    = eneC    = R*tempC/(gam - 1.0) + 0.5*(vxC*vxC + vyC*vyC + vzC*vzC);
	node[0].enthalC = enthalC = R*gam*tempC/(gam - 1.0) + 0.5*(vxC*vxC + vyC*vyC + vzC*vzC);

/* For the Current Time Step quantities */

	node[0].F[0] = dns*vx;          node[0].G[0] = dns*vy;          node[0].H[0] = dns*vz;
	node[0].F[1] = dns*vx*vx + prs; node[0].G[1] = dns*vy*vx;       node[0].H[1] = dns*vz*vx;
	node[0].F[2] = dns*vx*vy;       node[0].G[2] = dns*vy*vy + prs; node[0].H[2] = dns*vz*vy;
	node[0].F[3] = dns*vx*vz;       node[0].G[3] = dns*vy*vz;       node[0].H[3] = dns*vz*vz + prs;
	                                                    if(flag_2D) node[0].H[3] = 0.0;
	node[0].F[4] = dns*vx*enthal;   node[0].G[4] = dns*vy*enthal;   node[0].H[4] = dns*vz*enthal;

	node[0].U[0] = dns;
	node[0].U[1] = dns*vx;
	node[0].U[2] = dns*vy;
	node[0].U[3] = dns*vz;
	node[0].U[4] = dns*ene;

/* For the Predictor quantities */

	node[0].FP[0] = dnsP*vxP;            node[0].GP[0] = dnsP*vyP;            node[0].HP[0] = dnsP*vzP;
	node[0].FP[1] = dnsP*vxP*vxP + prsP; node[0].GP[1] = dnsP*vyP*vxP;        node[0].HP[1] = dnsP*vzP*vxP;
	node[0].FP[2] = dnsP*vxP*vyP;        node[0].GP[2] = dnsP*vyP*vyP + prsP; node[0].HP[2] = dnsP*vzP*vyP;
	node[0].FP[3] = dnsP*vxP*vzP;        node[0].GP[3] = dnsP*vyP*vzP;        node[0].HP[3] = dnsP*vzP*vzP + prsP;
	                                                              if(flag_2D) node[0].HP[3] = 0.0;
	node[0].FP[4] = dnsP*vxP*enthalP;    node[0].GP[4] = dnsP*vyP*enthalP;    node[0].HP[4] = dnsP*vzP*enthalP;

	node[0].UP[0] = dnsP;
	node[0].UP[1] = dnsP*vxP;
	node[0].UP[2] = dnsP*vyP;
	node[0].UP[3] = dnsP*vzP;
	node[0].UP[4] = dnsP*eneP;

/* For the Corrector quantities */

	node[0].FC[0] = dnsC*vxC;            node[0].GC[0] = dnsC*vyC;            node[0].HC[0] = dnsC*vzC;
	node[0].FC[1] = dnsC*vxC*vxC + prsC; node[0].GC[1] = dnsC*vyC*vxC;        node[0].HC[1] = dnsC*vzC*vxC;
	node[0].FC[2] = dnsC*vxC*vyC;        node[0].GC[2] = dnsC*vyC*vyC + prsC; node[0].HC[2] = dnsC*vzC*vyC;
	node[0].FC[3] = dnsC*vxC*vzC;        node[0].GC[3] = dnsC*vyC*vzC;        node[0].HC[3] = dnsC*vzC*vzC + prsC;
	                                                              if(flag_2D) node[0].HC[3] = 0.0;
	node[0].FC[4] = dnsC*vxC*enthalC;    node[0].GC[4] = dnsC*vyC*enthalC;    node[0].HC[4] = dnsC*vzC*enthalC;

	node[0].UC[0] = dnsC;
	node[0].UC[1] = dnsC*vxC;
	node[0].UC[2] = dnsC*vyC;
	node[0].UC[3] = dnsC*vzC;
	node[0].UC[4] = dnsC*eneC;

	return 1;
}


int applyBC_free (COM_NODE *node, COM_NODE *node1, COM_NODE *node2)
{
/* The first node above is the one lying on the boundary and "node1" 
   nade "node2" are the nearby nodes located in the opposite direction.  I
   made:

     applyBC_free (COM_NODE *node, COM_NODE *node1, COM_NODE *node2)

   a separate function because it makes everything more readable and I
   am hoping it won't be too inefficient compared to leaving these
   calculations within the loop where they are originally called.

             Updated 4/7/06
*/

	int i, j, k, dum, node_num;
	double dns, temp, prs, vx, vy, vz, enthal, ene;
	double dnsP, tempP, prsP, vxP, vyP, vzP, enthalP, eneP;
	double dnsC, tempC, prsC, vxC, vyC, vzC, enthalC, eneC;

/* For the Current Time Step quantities */

	node[0].dns    = dns    =  2.0*node1[0].dns - node2[0].dns;
	node[0].temp   = temp   =  2.0*node1[0].temp - node2[0].temp;
	node[0].prs    = prs    =  2.0*node1[0].prs - node2[0].prs;
	node[0].vx     = vx     =  2.0*node1[0].vx - node2[0].vx;
	node[0].vy     = vy     =  2.0*node1[0].vy -  node2[0].vy;
	node[0].vz     = vz     =  2.0*node1[0].vz -  node2[0].vz;
	if(flag_2D) vz = 0.0000;
	node[0].ene    = ene    =  2.0*node1[0].ene - node2[0].ene;
	node[0].enthal = enthal =  2.0*node1[0].enthal - node2[0].enthal;

/* For the Predictor quantities */

	node[0].dnsP    = dnsP    =  2.0*node1[0].dnsP - node2[0].dnsP;
	node[0].tempP   = tempP   =  2.0*node1[0].tempP - node2[0].tempP;
	node[0].prsP    = prsP    =  2.0*node1[0].prsP - node2[0].prsP;
	node[0].vxP     = vxP     =  2.0*node1[0].vxP - node2[0].vxP;
	node[0].vyP     = vyP     =  2.0*node1[0].vyP -  node2[0].vyP;
	node[0].vzP     = vzP     =  2.0*node1[0].vzP -  node2[0].vzP;
	if(flag_2D) vzP = 0.0000;
	node[0].eneP    = eneP    =  2.0*node1[0].eneP - node2[0].eneP;
	node[0].enthalP = enthalP =  2.0*node1[0].enthalP - node2[0].enthalP;

/* For the Corrector quantities */

	node[0].dnsC    = dnsC    =  2.0*node1[0].dnsC - node2[0].dnsC;
	node[0].tempC   = tempC   =  2.0*node1[0].tempC - node2[0].tempC;
	node[0].prsC    = prsC    =  2.0*node1[0].prsC - node2[0].prsC;
	node[0].vxC     = vxC     =  2.0*node1[0].vxC - node2[0].vxC;
	node[0].vyC     = vyC     =  2.0*node1[0].vyC -  node2[0].vyC;
	node[0].vzC     = vzC     =  2.0*node1[0].vzC -  node2[0].vzC;
	if(flag_2D) vzC = 0.0000;
	node[0].eneC    = eneC    =  2.0*node1[0].eneC - node2[0].eneC;
	node[0].enthalC = enthalC =  2.0*node1[0].enthalC - node2[0].enthalC;

/* For the Current Time Step quantities */

	node[0].F[0] = dns*vx;          node[0].G[0] = dns*vy;          node[0].H[0] = dns*vz;
	node[0].F[1] = dns*vx*vx + prs; node[0].G[1] = dns*vy*vx;       node[0].H[1] = dns*vz*vx;
	node[0].F[2] = dns*vx*vy;       node[0].G[2] = dns*vy*vy + prs; node[0].H[2] = dns*vz*vy;
	node[0].F[3] = dns*vx*vz;       node[0].G[3] = dns*vy*vz;       node[0].H[3] = dns*vz*vz + prs;
	                                                    if(flag_2D) node[0].H[3] = 0.0;
	node[0].F[4] = dns*vx*enthal;   node[0].G[4] = dns*vy*enthal;   node[0].H[4] = dns*vz*enthal;

	node[0].U[0] = dns;
	node[0].U[1] = dns*vx;
	node[0].U[2] = dns*vy;
	node[0].U[3] = dns*vz;
	node[0].U[4] = dns*ene;

/* For the Predictor quantities */

	node[0].FP[0] = dnsP*vxP;            node[0].GP[0] = dnsP*vyP;            node[0].HP[0] = dnsP*vzP;
	node[0].FP[1] = dnsP*vxP*vxP + prsP; node[0].GP[1] = dnsP*vyP*vxP;        node[0].HP[1] = dnsP*vzP*vxP;
	node[0].FP[2] = dnsP*vxP*vyP;        node[0].GP[2] = dnsP*vyP*vyP + prsP; node[0].HP[2] = dnsP*vzP*vyP;
	node[0].FP[3] = dnsP*vxP*vzP;        node[0].GP[3] = dnsP*vyP*vzP;        node[0].HP[3] = dnsP*vzP*vzP + prsP;
	                                                              if(flag_2D) node[0].HP[3] = 0.0;
	node[0].FP[4] = dnsP*vxP*enthalP;    node[0].GP[4] = dnsP*vyP*enthalP;    node[0].HP[4] = dnsP*vzP*enthalP;

	node[0].UP[0] = dnsP;
	node[0].UP[1] = dnsP*vxP;
	node[0].UP[2] = dnsP*vyP;
	node[0].UP[3] = dnsP*vzP;
	node[0].UP[4] = dnsP*eneP;

/* For the Corrector quantities */

	node[0].FC[0] = dnsC*vxC;            node[0].GC[0] = dnsC*vyC;            node[0].HC[0] = dnsC*vzC;
	node[0].FC[1] = dnsC*vxC*vxC + prsC; node[0].GC[1] = dnsC*vyC*vxC;        node[0].HC[1] = dnsC*vzC*vxC;
	node[0].FC[2] = dnsC*vxC*vyC;        node[0].GC[2] = dnsC*vyC*vyC + prsC; node[0].HC[2] = dnsC*vzC*vyC;
	node[0].FC[3] = dnsC*vxC*vzC;        node[0].GC[3] = dnsC*vyC*vzC;        node[0].HC[3] = dnsC*vzC*vzC + prsC;
	                                                              if(flag_2D) node[0].HC[3] = 0.0;
	node[0].FC[4] = dnsC*vxC*enthalC;    node[0].GC[4] = dnsC*vyC*enthalC;    node[0].HC[4] = dnsC*vzC*enthalC;


	node[0].UC[0] = dnsC;
	node[0].UC[1] = dnsC*vxC;
	node[0].UC[2] = dnsC*vyC;
	node[0].UC[3] = dnsC*vzC;
	node[0].UC[4] = dnsC*eneC;

	return 1;
}




int applyBC_noslip( BOUND bc, MATL *matl, COM_NODE *node )
{
/* The no slip BC is applied.  Because I am using a staggard grid, there are many nodes
   which do not fall on the edge nodes of the box.  So for some nodes, I have to use
   extrapolation.

   The boundary condtions for noslip also apply to some pressure nodes.  The physical condition is
   that the pressure gradient normal to the surface is zero.  This is given on page 340-341,
   section 17.1.3 in:

     Fletcher, Clive A. J., Computational Techniques for Fluid Dynamics: Vol I, 2nd
        Ed., Springer-Verlag, Berlin, 1991.

   I do not apply it exactly as givin by Fletcher because what I currently do gives better
   results.  Overall, this pressure condition slightly improves the results.

   All of these boundary conditions are mixed together below, except for setting the velocity
   to zero inside the body.
   
                     Updated 4/6/08
*/

	int i, j, k, dum, node_num, check;
	int ce, xm1, xm2, xp1, xp2, ym1, ym2, yp1, yp2, xm1ym1, xm1yp1, xm2ym1, xp1ym1, xp1yp1;

#if 1
/* Set nodes internal to the body to zero */

	for( dum = 0; dum < bc.num_vel0[0].x; ++dum)
	{
		node_num = bc.vel0.x[dum];
		i = node[node_num].i;
		j = node[node_num].j;
		ce =  ny*i + j;
		node[ce].vx = 0.0;
		node[ce].vxP = 0.0;
		node[ce].vxC = 0.0;

		check = setFGHU ( matl, node, ce);
		if(!check) printf( " Problems with setFGHU \n");
	}

	for( dum = 0; dum < bc.num_vel0[0].y; ++dum)
	{
		node_num = bc.vel0.y[dum];
		i = node[node_num].i;
		j = node[node_num].j;
		ce =  ny*i + j;
		node[ce].vy = 0.0;
		node[ce].vyP = 0.0;
		node[ce].vyC = 0.0;

		check = setFGHU ( matl, node, ce);
		if(!check) printf( " Problems with setFGHU \n");
	}
#endif

	if( bc.num_noslp[0].xp > 0)
	{
/* noslip in x - */

	    dum = 0;
	    node_num = bc.noslp.xp[dum];
	    i = node[node_num].i;
	    j = node[node_num].j;
	    ce  = ny*i + j;
	    xm1 = ny*(i-1) + j;
	    xp1 = ny*(i+1) + j;

	    node[ce].vx = 0.0;     /* For the noslip vx velocities */
	    node[ce].vxP = 0.0;
	    node[ce].vxC = 0.0;

	    node[ce].vy  = -node[xp1].vy/3.0;   /* For the vy velocities immediately to the left */
	    node[ce].vyP = -node[xp1].vyP/3.0;   /* of the noslip velocities */
	    node[ce].vyC = -node[xp1].vyC/3.0;

	    node[ce].prs = node[xp1].prs;   /* For the pressure immediately to the left */
	    node[ce].prsP = node[xp1].prsP;   /* of the noslip velocities */
	    node[ce].prsC = node[xp1].prsC;

	    check = setFGHU ( matl, node, ce);
	    if(!check) printf( " Problems with setFGHU \n");

	    for( dum = 1; dum < bc.num_noslp[0].xp-1; ++dum)
	    {
	       node_num = bc.noslp.xp[dum];
	       i = node[node_num].i;
	       j = node[node_num].j;
	       ce  = ny*i + j;
	       xm1 = ny*(i-1) + j;
	       xp1 = ny*(i+1) + j;

	       node[ce].vx = 0.0;
	       node[ce].vxP = 0.0;
	       node[ce].vxC = 0.0;

	       node[ce].vy  = -node[xp1].vy/3.0;   /* For the vy velocities immediately to the left */
	       node[ce].vyP = -node[xp1].vyP/3.0;   /* of the noslip velocities */
	       node[ce].vyC = -node[xp1].vyC/3.0;

	       node[ce].prs = node[xp1].prs;   /* For the pressure immediately to the left */
	       node[ce].prsP = node[xp1].prsP;   /* of the noslip velocities */
	       node[ce].prsC = node[xp1].prsC;

	       check = setFGHU ( matl, node, ce);
	       if(!check) printf( " Problems with setFGHU \n");
	    }

	    dum = bc.num_noslp[0].xp-1;
	    node_num = bc.noslp.xp[dum];
	    i = node[node_num].i;
	    j = node[node_num].j;
	    ce  = ny*i + j;
	    xm1 = ny*(i-1) + j;
	    xp1 = ny*(i+1) + j;

	    node[ce].vx = 0.0;
	    node[ce].vxP = 0.0;
	    node[ce].vxC = 0.0;

	    node[ce].prs = node[xp1].prs;   /* For the pressure immediately to the left */
	    node[ce].prsP = node[xp1].prsP;   /* of the noslip velocities */
	    node[ce].prsC = node[xp1].prsC;

	    check = setFGHU ( matl, node, ce);
	    if(!check) printf( " Problems with setFGHU \n");
	}

	if( bc.num_noslp[0].xm > 0)
	{

/* noslip in x + */

	    dum = 0;
	    node_num = bc.noslp.xm[dum];
	    i = node[node_num].i;
	    j = node[node_num].j;
	    ce  = ny*i + j;
	    xm1 = ny*(i-1) + j;
	    xp1 = ny*(i+1) + j;
	    xp2 = ny*(i+2) + j;

	    node[ce].vx = 0.0;     /* For the noslip vx velocities */
	    node[ce].vxP = 0.0;
	    node[ce].vxC = 0.0;

	    node[xp1].vy  = -node[ce].vy/3.0;   /* For the vy velocities immediately to the right */
	    node[xp1].vyP = -node[ce].vyP/3.0;   /* of the noslip velocities */
	    node[xp1].vyC = -node[ce].vyC/3.0;

	    node[xp1].prs = node[ce].prs;   /* For the pressure immediately to the right */
	    node[xp1].prsP = node[ce].prsP;   /* of the noslip velocities */
	    node[xp1].prsC = node[ce].prsC;

	    check = setFGHU ( matl, node, ce);
	    if(!check) printf( " Problems with setFGHU \n");

	    for( dum = 1; dum < bc.num_noslp[0].xm-1; ++dum)
	    {
	       node_num = bc.noslp.xm[dum];
	       i = node[node_num].i;
	       j = node[node_num].j;
	       ce  = ny*i + j;
	       xm1 = ny*(i-1) + j;
	       xp1 = ny*(i+1) + j;
	       xp2 = ny*(i+2) + j;

	       node[ce].vx = 0.0;
	       node[ce].vxP = 0.0;
	       node[ce].vxC = 0.0;

	       node[xp1].vy  = -node[ce].vy/3.0;   /* For the vy velocities immediately to the right */
	       node[xp1].vyP = -node[ce].vyP/3.0;   /* of the noslip velocities */
	       node[xp1].vyC = -node[ce].vyC/3.0;

	       node[xp1].prs = node[ce].prs;   /* For the pressure immediately to the right */
	       node[xp1].prsP = node[ce].prsP;   /* of the noslip velocities */
	       node[xp1].prsC = node[ce].prsC;

	       check = setFGHU ( matl, node, ce);
	       if(!check) printf( " Problems with setFGHU \n");
	    }

	    dum = bc.num_noslp[0].xm-1;
	    node_num = bc.noslp.xm[dum];
	    i = node[node_num].i;
	    j = node[node_num].j;
	    ce  = ny*i + j;
	    xm1 = ny*(i-1) + j;
	    xp1 = ny*(i+1) + j;
	    xp2 = ny*(i+2) + j;

	    node[ce].vx = 0.0;
	    node[ce].vxP = 0.0;
	    node[ce].vxC = 0.0;

	    node[xp1].prs = node[ce].prs;   /* For the pressure immediately to the right */
	    node[xp1].prsP = node[ce].prsP;   /* of the noslip velocities */
	    node[xp1].prsC = node[ce].prsC;

	    check = setFGHU ( matl, node, ce);
	    if(!check) printf( " Problems with setFGHU \n");
	}

	if( bc.num_noslp[0].yp > 0)
	{

/* noslip in y - */

	    dum = 0;
	    node_num = bc.noslp.yp[dum];
	    i = node[node_num].i;
	    j = node[node_num].j;
	    ce  = ny*i + j;
	    ym1 = ny*i + j-1;
	    yp1 = ny*i + j+1;

	    node[ce].vy = 0.0;     /* For the noslip vx velocities */
	    node[ce].vyP = 0.0;
	    node[ce].vyC = 0.0;

	    node[ce].vx  = -node[yp1].vx/3.0;   /* For the vy velocities immediately bolow the top edge */
	    node[ce].vxP = -node[yp1].vxP/3.0;   /* of the noslip velocities */
	    node[ce].vxC = -node[yp1].vxC/3.0;

	    node[ce].prs = node[yp1].prs;   /* For the pressure immediately bolow the top edge */
	    node[ce].prsP = node[yp1].prsP;   /* of the noslip velocities */
	    node[ce].prsC = node[yp1].prsC;

	    check = setFGHU ( matl, node, ce);
	    if(!check) printf( " Problems with setFGHU \n");

	    for( dum = 1; dum < bc.num_noslp[0].yp-1; ++dum)
	    {
	       node_num = bc.noslp.yp[dum];
	       i = node[node_num].i;
	       j = node[node_num].j;
	       ce  = ny*i + j;
	       ym1 = ny*i + j-1;
	       yp1 = ny*i + j+1;

	       node[ce].vy = 0.0;
	       node[ce].vyP = 0.0;
	       node[ce].vyC = 0.0;

	       node[ce].vx  = -node[yp1].vx/3.0;   /* For the vy velocities immediately bolow the top edge */
	       node[ce].vxP = -node[yp1].vxP/3.0;   /* of the noslip velocities */
	       node[ce].vxC = -node[yp1].vxC/3.0;

	       node[ce].prs = node[yp1].prs;   /* For the pressure immediately bolow the top edge */
	       node[ce].prsP = node[yp1].prsP;   /* of the noslip velocities */
	       node[ce].prsC = node[yp1].prsC;

	       check = setFGHU ( matl, node, ce);
	       if(!check) printf( " Problems with setFGHU \n");
	    }

	    dum = bc.num_noslp[0].yp-1;
	    node_num = bc.noslp.yp[dum];
	    i = node[node_num].i;
	    j = node[node_num].j;
	    ce  = ny*i + j;
	    ym1 = ny*i + j-1;
	    yp1 = ny*i + j+1;

	    node[ce].vy = 0.0;
	    node[ce].vyP = 0.0;
	    node[ce].vyC = 0.0;

	    node[ce].prs = node[yp1].prs;   /* For the pressure immediately bolow the top edge */
	    node[ce].prsP = node[yp1].prsP;   /* of the noslip velocities */
	    node[ce].prsC = node[yp1].prsC;

	    check = setFGHU ( matl, node, ce);
	    if(!check) printf( " Problems with setFGHU \n");

	}
	if( bc.num_noslp[0].ym > 0)
	{

/*     noslip in y + */

	    dum = 0;
	    node_num = bc.noslp.ym[dum];
	    i = node[node_num].i;
	    j = node[node_num].j;
	    ce  = ny*i + j;
	    ym1 = ny*i + j-1;
	    yp1 = ny*i + j+1;
	    yp2 = ny*i + j+2;

	    node[ce].vy = 0.0;     /* For the noslip vx velocities */
	    node[ce].vyP = 0.0;
	    node[ce].vyC = 0.0;

	    node[yp1].vx  = -node[ce].vx/3.0;   /* For the vy velocities immediately above the bottom edge */
	    node[yp1].vxP = -node[ce].vxP/3.0;   /* of the noslip velocities */
	    node[yp1].vxC = -node[ce].vxC/3.0;

	    node[yp1].prs = node[ce].prs;   /* For the pressure immediately above the bottom edge */
	    node[yp1].prsP = node[ce].prsP;
	    node[yp1].prsC = node[ce].prsC;

	    check = setFGHU ( matl, node, ce);
	    if(!check) printf( " Problems with setFGHU \n");

	    for( dum = 1; dum < bc.num_noslp[0].ym-1; ++dum)
	    {
	       node_num = bc.noslp.ym[dum];
	       i = node[node_num].i;
	       j = node[node_num].j;
	       ce  = ny*i + j;
	       ym1 = ny*i + j-1;
	       yp1 = ny*i + j+1;
	       yp2 = ny*i + j+2;

	       node[ce].vy = 0.0;
	       node[ce].vyP = 0.0;
	       node[ce].vyC = 0.0;

	       node[yp1].vx  = -node[ce].vx/3.0;   /* For the vy velocities immediately above the bottom edge */
	       node[yp1].vxP = -node[ce].vxP/3.0;   /* of the noslip velocities */
	       node[yp1].vxC = -node[ce].vxC/3.0;

	       node[yp1].prs = node[ce].prs;   /* For the pressure immediately above the bottom edge */
	       node[yp1].prsP = node[ce].prsP;   /* of the noslip velocities */
	       node[yp1].prsC = node[ce].prsC;

	       check = setFGHU ( matl, node, ce);
	       if(!check) printf( " Problems with setFGHU \n");
	    }

	    dum = bc.num_noslp[0].ym-1;
	    node_num = bc.noslp.ym[dum];
	    i = node[node_num].i;
	    j = node[node_num].j;
	    ce  = ny*i + j;
	    ym1 = ny*i + j-1;
	    yp1 = ny*i + j+1;
	    yp2 = ny*i + j+2;

	    node[ce].vy = 0.0;
	    node[ce].vyP = 0.0;
	    node[ce].vyC = 0.0;

	    node[yp1].prs = node[ce].prs;   /* For the pressure immediately above the bottom edge */
	    node[yp1].prsP = node[ce].prsP;
	    node[yp1].prsC = node[ce].prsC;

	    check = setFGHU ( matl, node, ce);
	    if(!check) printf( " Problems with setFGHU \n");
	}


/* These are for the corner nodes where there is overlap between noslip in x and y. */

	for( dum = 0; dum < bc.num_noslp_prs[0].xpyp; ++dum )
	{
	    node_num = bc.noslp_prs.xpyp[0];
	    i = node[node_num].i;
	    j = node[node_num].j;
	    ce  = ny*i + j;
	    xm1 = ny*(i-1) + j;
	    xp1 = ny*(i+1) + j;
	    ym1 = ny*i + j-1;
	    yp1 = ny*i + j+1;
	    node[ce].prs = .5*(node[xp1].prs + node[yp1].prs);
	    node[ce].prsP = .5*(node[xp1].prsP + node[yp1].prsP);
	    node[ce].prsC = .5*(node[xp1].prsC + node[yp1].prsC);

	    check = setFGHU ( matl, node, ce);
	    if(!check) printf( " Problems with setFGHU \n");
	}
	for( dum = 0; dum < bc.num_noslp_prs[0].xpym; ++dum )
	{
	   node_num = bc.noslp_prs.xpym[0];
	   i = node[node_num].i;
	   j = node[node_num].j;
	   ce  = ny*i + j;
	   xm1 = ny*(i-1) + j;
	   xp1 = ny*(i+1) + j;
	   ym1 = ny*i + j-1;
	   yp1 = ny*i + j+1;
	   node[ce].prs = .5*(node[xp1].prs + node[ym1].prs);
	   node[ce].prsP = .5*(node[xp1].prsP + node[ym1].prsP);
	   node[ce].prsC = .5*(node[xp1].prsC + node[ym1].prsC);
	}
	for( dum = 0; dum < bc.num_noslp_prs[0].xmyp; ++dum )
	{
	   node_num = bc.noslp_prs.xmyp[0];
	   i = node[node_num].i;
	   j = node[node_num].j;
	   ce  = ny*i + j;
	   xm1 = ny*(i-1) + j;
	   xp1 = ny*(i+1) + j;
	   ym1 = ny*i + j-1;
	   yp1 = ny*i + j+1;
	   node[ce].prs = .5*(node[xm1].prs + node[yp1].prs);
	   node[ce].prsP = .5*(node[xm1].prsP + node[yp1].prsP);
	   node[ce].prsC = .5*(node[xm1].prsC + node[yp1].prsC);
	}
	for( dum = 0; dum < bc.num_noslp_prs[0].xmym; ++dum )
	{
	   node_num = bc.noslp_prs.xmym[0];
	   i = node[node_num].i;
	   j = node[node_num].j;
	   ce  = ny*i + j;
	   xm1 = ny*(i-1) + j;
	   xp1 = ny*(i+1) + j;
	   ym1 = ny*i + j-1;
	   yp1 = ny*i + j+1;
	   node[ce].prs = .5*(node[xm1].prs + node[ym1].prs);
	   node[ce].prsP = .5*(node[xm1].prsP + node[ym1].prsP);
	   node[ce].prsC = .5*(node[xm1].prsC + node[ym1].prsC);
	}

	return 1;

}

int setFGHU (MATL *matl, COM_NODE *node, int ce)
{
/*
   This subrountine stes F, G, H, U as well as their predictor/corrector
   quantities.

             Updated 2/2/06
*/

	int i, j, k, dum, node_num;
	double dns, temp, prs, vx, vy, vz, enthal, ene;
	double dnsP, tempP, prsP, vxP, vyP, vzP, enthalP, eneP;
	double dnsC, tempC, prsC, vxC, vyC, vzC, enthalC, eneC;
	double gam, R;

	gam = matl[0].gam;
	R = matl[0].R;

/* For the Current Time Step quantities */

	dns    =  node[ce].dns;
	temp   =  node[ce].temp;
	prs    =  node[ce].prs;
	vx     =  node[ce].vx;
	vy     =  node[ce].vy;
	vz     =  node[ce].vz;
	if(flag_2D) vz = 0.0000;
	ene    = R*temp/(gam - 1.0) + 0.5*(vx*vx + vy*vy + vz*vz);
	enthal = R*gam*temp/(gam - 1.0) + 0.5*(vx*vx + vy*vy + vz*vz);

/* For the Predictor quantities */

	dnsP    =  node[ce].dnsP;
	tempP   =  node[ce].tempP;
	prsP    =  node[ce].prsP;
	vxP     =  node[ce].vxP;
	vyP     =  node[ce].vyP;
	vzP     =  node[ce].vzP;
	if(flag_2D) vzP = 0.0000;
	eneP    = R*tempP/(gam - 1.0) + 0.5*(vxP*vxP + vyP*vyP + vzP*vzP);
	enthalP = R*gam*tempP/(gam - 1.0) + 0.5*(vxP*vxP + vyP*vyP + vzP*vzP);

/* For the Corrector quantities */

	dnsC    =  node[ce].dnsC;
	tempC   =  node[ce].tempC;
	prsC    =  node[ce].prsC;
	vxC     =  node[ce].vxC;
	vyC     =  node[ce].vyC;
	vzC     =  node[ce].vzC;
	if(flag_2D) vzC = 0.0000;
	eneC    = R*tempC/(gam - 1.0) + 0.5*(vxC*vxC + vyC*vyC + vzC*vzC);
	enthalC = R*gam*tempC/(gam - 1.0) + 0.5*(vxC*vxC + vyC*vyC + vzC*vzC);

/* For the Current Time Step quantities */

	node[ce].F[0] = dns*vx;          node[ce].G[0] = dns*vy;          node[ce].H[0] = dns*vz;
	node[ce].F[1] = dns*vx*vx + prs; node[ce].G[1] = dns*vy*vx;       node[ce].H[1] = dns*vz*vx;
	node[ce].F[2] = dns*vx*vy;       node[ce].G[2] = dns*vy*vy + prs; node[ce].H[2] = dns*vz*vy;
	node[ce].F[3] = dns*vx*vz;       node[ce].G[3] = dns*vy*vz;       node[ce].H[3] = dns*vz*vz + prs;
	                                                    if(flag_2D) node[ce].H[3] = 0.0;
	node[ce].F[4] = dns*vx*enthal;   node[ce].G[4] = dns*vy*enthal;   node[ce].H[4] = dns*vz*enthal;

	node[ce].U[0] = dns;
	node[ce].U[1] = dns*vx;
	node[ce].U[2] = dns*vy;
	node[ce].U[3] = dns*vz;
	node[ce].U[4] = dns*ene;

/* For the Predictor quantities */

	node[ce].FP[0] = dnsP*vxP;            node[ce].GP[0] = dnsP*vyP;            node[ce].HP[0] = dnsP*vzP;
	node[ce].FP[1] = dnsP*vxP*vxP + prsP; node[ce].GP[1] = dnsP*vyP*vxP;        node[ce].HP[1] = dnsP*vzP*vxP;
	node[ce].FP[2] = dnsP*vxP*vyP;        node[ce].GP[2] = dnsP*vyP*vyP + prsP; node[ce].HP[2] = dnsP*vzP*vyP;
	node[ce].FP[3] = dnsP*vxP*vzP;        node[ce].GP[3] = dnsP*vyP*vzP;        node[ce].HP[3] = dnsP*vzP*vzP + prsP;
	                                                              if(flag_2D) node[ce].HP[3] = 0.0;
	node[ce].FP[4] = dnsP*vxP*enthalP;    node[ce].GP[4] = dnsP*vyP*enthalP;    node[ce].HP[4] = dnsP*vzP*enthalP;

	node[ce].UP[0] = dnsP;
	node[ce].UP[1] = dnsP*vxP;
	node[ce].UP[2] = dnsP*vyP;
	node[ce].UP[3] = dnsP*vzP;
	node[ce].UP[4] = dnsP*eneP;

/* For the Corrector quantities */

	node[ce].FC[0] = dnsC*vxC;            node[ce].GC[0] = dnsC*vyC;            node[ce].HC[0] = dnsC*vzC;
	node[ce].FC[1] = dnsC*vxC*vxC + prsC; node[ce].GC[1] = dnsC*vyC*vxC;        node[ce].HC[1] = dnsC*vzC*vxC;
	node[ce].FC[2] = dnsC*vxC*vyC;        node[ce].GC[2] = dnsC*vyC*vyC + prsC; node[ce].HC[2] = dnsC*vzC*vyC;
	node[ce].FC[3] = dnsC*vxC*vzC;        node[ce].GC[3] = dnsC*vyC*vzC;        node[ce].HC[3] = dnsC*vzC*vzC + prsC;
	                                                              if(flag_2D) node[ce].HC[3] = 0.0;
	node[ce].FC[4] = dnsC*vxC*enthalC;    node[ce].GC[4] = dnsC*vyC*enthalC;    node[ce].HC[4] = dnsC*vzC*enthalC;

	node[ce].UC[0] = dnsC;
	node[ce].UC[1] = dnsC*vxC;
	node[ce].UC[2] = dnsC*vyC;
	node[ce].UC[3] = dnsC*vzC;
	node[ce].UC[4] = dnsC*eneC;

	return 1;
}

