/* This program performs the corrector step for the program which
   uses the finite difference method to calculate a compressible CFD flow field.
   The algorithm for the code below based on the Forney-Lax-Wendroff 2-step. 
   The main reference is:

Forney, L. J., Particle Impaction on Axially Symmetric Supersonic Flow, Journal of
   Aerosol Science and Technology 15, page 49, 1991.

It is similar to the MacCormack 2-Step seen on page 261 of:

Hirsch, C., Numerical Computation of Internal and External Flows Vol. 2: Computational
   Methods for Inviscid and Viscous Flows, John Wiley & Sons, Chichester, 1994. 

This same reference on page 135 gives a good discussion of the conservative form of
Euler's Equations.

I have extrapolated the 2-D version to 3-D.

                     Updated 2/2/06

    SLFCFD source file
    Version:  1.1
    Copyright (C) 2005, 2006, 2007, 2008  San Le

    The source code contained in this file is released under the
    terms of the GNU Library General Public License.

*/

#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "cfdconst.h"
#include "cfdstruct.h"

extern dof, numel, numnp, sof, nx, ny, nz;
extern int flag_2D, flag_3D;
extern double dt, dx, dy, dz;

int ForneyLaxWendroffCorrector ( BOUND bc, MATL *matl, COM_NODE *node)
{
/*

  Use the key below to understand the indexing.

         ce: node (i, j, k )
         xm1: node (i - 1, j, k)
         xp1: node (i + 1, j, k)
         ym1: node (i, j - 1, k)
         yp1: node (i, j + 1, k)
         zm1: node (i, j, k - 1)
         zp1: node (i, j, k + 1)
*/

	int i, j, k, i2, dum, node_num;
	int ce, xp1, xm1, yp1, ym1, zp1, zm1;
	double dns, prs, temp, enthal, ene, vx, vy, vz;
	double dnsP, prsP, tempP, enthalP, eneP, vxP, vyP, vzP;
	double dnsC, prsC, tempC, enthalC, eneC, vxC, vyC, vzC;
	double gam, R;

	double FceP[5], Fxm1P[5], Fxp1P[5], Fym1P[5], Fyp1P[5], Fzm1P[5], Fzp1P[5];
	double GceP[5], Gxm1P[5], Gxp1P[5], Gym1P[5], Gyp1P[5], Gzm1P[5], Gzp1P[5];
	double HceP[5], Hxm1P[5], Hxp1P[5], Hym1P[5], Hyp1P[5], Hzm1P[5], Hzp1P[5];
	double Uce[5], Uxm1[5], Uxp1[5], Uym1[5], Uyp1[5], Uzm1[5], Uzp1[5];
	double UC[5];

	double taux = dt/dx;
	double tauy = dt/dy;
	double tauz = dt/dz;
	double fdum;

	gam = matl[0].gam;
	R = matl[0].R;

	if(flag_3D)
	{
	    for ( dum=0; dum < bc.num_none[0]; dum++)
	    {
		node_num = bc.none[dum];
		i = node[node_num].i;
		j = node[node_num].j;
		k = node[node_num].k;
		ce = nz*ny*i + nz*j + k;
		xm1 = nz*ny*(i-1) + nz*j + k;
		xp1 = nz*ny*(i+1) + nz*j + k;
		ym1 = nz*ny*i + nz*(j-1) + k;
		yp1 = nz*ny*i + nz*(j+1) + k;
		zm1 = nz*ny*i + nz*j + (k-1);
		zp1 = nz*ny*i + nz*j + (k+1);

		//printf("rrrr %4d %4d %4d vz\n", i, j, k);

		for ( i2 = 0; i2 < 5; i2++)
		{
/*     Based on the calculation for UC[i2] below, I probably don't need all the positions for
       F, G, H, and U to be calculated.  Remember to comment them out later.
*/
#if 0
		    FceP[i2]  = node[ce].FP[i2];   GceP[i2]  = node[ce].GP[i2];   HceP[i2]  = node[ce].HP[i2]; 
		    Fxm1P[i2] = node[xm1].FP[i2];  Gxm1P[i2] = node[xm1].GP[i2];  Hxm1P[i2] = node[xm1].HP[i2];
		    Fxp1P[i2] = node[xp1].FP[i2];  Gxp1P[i2] = node[xp1].GP[i2];  Hxp1P[i2] = node[xp1].HP[i2];
		    Fym1P[i2] = node[ym1].FP[i2];  Gym1P[i2] = node[ym1].GP[i2];  Hym1P[i2] = node[ym1].HP[i2];
		    Fyp1P[i2] = node[yp1].FP[i2];  Gyp1P[i2] = node[yp1].GP[i2];  Hyp1P[i2] = node[yp1].HP[i2];
		    Fzm1P[i2] = node[zm1].FP[i2];  Gzm1P[i2] = node[zm1].GP[i2];  Hzm1P[i2] = node[zm1].HP[i2];
		    Fzp1P[i2] = node[zp1].FP[i2];  Gzp1P[i2] = node[zp1].GP[i2];  Hzp1P[i2] = node[zp1].HP[i2];

		    Uce[i2]  = node[ce].U[i2]; 
		    Uxm1[i2] = node[xm1].U[i2];
		    Uxp1[i2] = node[xp1].U[i2];
		    Uym1[i2] = node[ym1].U[i2];
		    Uyp1[i2] = node[yp1].U[i2];
		    Uzm1[i2] = node[zm1].U[i2];
		    Uzp1[i2] = node[zp1].U[i2];
#endif
		    FceP[i2]  = node[ce].FP[i2];   GceP[i2]  = node[ce].GP[i2];   HceP[i2]  = node[ce].HP[i2]; 
		    Fxm1P[i2] = node[xm1].FP[i2];
		                                   Gym1P[i2] = node[ym1].GP[i2];
		                                                                  Hzm1P[i2] = node[zm1].HP[i2];
		    Uce[i2]  = node[ce].U[i2]; 

		    UC[i2] = Uce[i2] - taux*(FceP[i2] - Fxm1P[i2]) -
		            tauy*(GceP[i2] - Gym1P[i2]) - tauz*(HceP[i2] - Hzm1P[i2]);
		    node[ce].UC[i2] = 0.5*(node[ce].UP[i2] + UC[i2]);
		    //printf("ccccc vz %16.8e %16.8e %16.8e %16.8e %16.8e %16.8e %16.8e\n",
		    //   UC[i2], Uxm1P[i2], Uxp1P[i2], Uym1P[i2], Uyp1P[i2], Uzm1P[i2], Uzp1P[i2]);

		}

/*     San: Note that unlike UP, it is not necessary to have UC, and to simply replace
       all instances of *C quantities with *(e.g. dns instead of dnsC, F instead of FC, etc.).
       The reason I keep this extra step with the *C quantities is because the current
       method of numerical integration( Lax-Forney 2 step) only requires U at the current
       node for U.  If it had required U at surrounding nodes(xm1, xp1, ym1, yp1, zm1, zp1),
       I would definitely need a UC to act as a temporary place holder.

       Also note that if you choose to go without *C quantities, then you need to remove
       the Reset step in Model::solve.
*/

		node[ce].dnsC    = dnsC    = node[ce].UC[0];
		node[ce].vxC     = vxC     = node[ce].UC[1]/dnsC;
		node[ce].vyC     = vyC     = node[ce].UC[2]/dnsC;
		node[ce].vzC     = vzC     = node[ce].UC[3]/dnsC;
		node[ce].eneC    = eneC    = node[ce].UC[4]/dnsC;
		node[ce].tempC   = tempC   = (1.0/R)*(gam - 1.0) * ( eneC - 0.5*(vxC*vxC + vyC*vyC + vzC*vzC) );
		node[ce].enthalC = enthalC = R*gam*tempC/(gam - 1.0) + 0.5*(vxC*vxC + vyC*vyC + vzC*vzC);
		node[ce].prsC    = prsC    = R*dnsC*tempC;

		/* set F Corrector */                  /* set G Corrector */                  /* set H Corrector */

		node[ce].FC[0] = dnsC*vxC;             node[ce].GC[0] = dnsC*vyC;             node[ce].HC[0] = dnsC*vzC;
		node[ce].FC[1] = dnsC*vxC*vxC + prsC;  node[ce].GC[1] = dnsC*vyC*vxC;         node[ce].HC[1] = dnsC*vzC*vxC;
		node[ce].FC[2] = dnsC*vxC*vyC;         node[ce].GC[2] = dnsC*vyC*vyC + prsC;  node[ce].HC[2] = dnsC*vzC*vyC;
		node[ce].FC[3] = dnsC*vxC*vzC;         node[ce].GC[3] = dnsC*vyC*vzC;         node[ce].HC[3] = dnsC*vzC*vzC + prsC;
		node[ce].FC[4] = dnsC*vxC*enthalC;     node[ce].GC[4] = dnsC*vyC*enthalC;     node[ce].HC[4] = dnsC*vzC*enthalC;
	    }
	}

	if(flag_2D)
	{
	    for ( dum=0; dum < bc.num_none[0]; dum++)
	    {
		node_num = bc.none[dum];
		i = node[node_num].i;
		j = node[node_num].j;
		ce = ny*i + j;
		xm1 = ny*(i-1) + j;
		xp1 = ny*(i+1) + j;
		ym1 = ny*i + (j-1);
		yp1 = ny*i + (j+1);

		//printf("rrrr %4d %4d %4d vz\n", i, j, k);

		for ( i2 = 0; i2 < 5; i2++)
		{
/*     Based on the calculation for UC[i2] below, I probably don't need all the positions for
       F, G, H, and U to be calculated.  Remember to comment them out later.
*/
#if 0
		    FceP[i2]  = node[ce].FP[i2];   GceP[i2]  = node[ce].GP[i2];
		    Fxm1P[i2] = node[xm1].FP[i2];  Gxm1P[i2] = node[xm1].GP[i2];
		    Fxp1P[i2] = node[xp1].FP[i2];  Gxp1P[i2] = node[xp1].GP[i2];
		    Fym1P[i2] = node[ym1].FP[i2];  Gym1P[i2] = node[ym1].GP[i2];
		    Fyp1P[i2] = node[yp1].FP[i2];  Gyp1P[i2] = node[yp1].GP[i2];

		    Uce[i2]  = node[ce].U[i2]; 
		    Uxm1[i2] = node[xm1].U[i2];
		    Uxp1[i2] = node[xp1].U[i2];
		    Uym1[i2] = node[ym1].U[i2];
		    Uyp1[i2] = node[yp1].U[i2];
#endif
		    FceP[i2]  = node[ce].FP[i2];   GceP[i2]  = node[ce].GP[i2];
		    Fxm1P[i2] = node[xm1].FP[i2];
		                                   Gym1P[i2] = node[ym1].GP[i2];
		    Uce[i2]  = node[ce].U[i2]; 

		    UC[i2] = Uce[i2] - taux*(FceP[i2] - Fxm1P[i2]) -
		            tauy*(GceP[i2] - Gym1P[i2]);
		    node[ce].UC[i2] = 0.5*(node[ce].UP[i2] + UC[i2]);
		    //printf("ccccc vz %16.8e %16.8e %16.8e %16.8e %16.8e %16.8e %16.8e\n",
		    //   UC[i2], Uxm1P[i2], Uxp1P[i2], Uym1P[i2], Uyp1P[i2], Uzm1P[i2], Uzp1P[i2]);

		}

/*     San: Note that unlike UP, it is not necessary to have UC, and to simply replace
       all instances of *C quantities with *(e.g. dns instead of dnsC, F instead of FC, etc.).
       The reason I keep this extra step with the *C quantities is because the current
       method of numerical integration( Lax-Forney 2 step) only requires U at the current
       node for U.  If it had required U at surrounding nodes(xm1, xp1, ym1, yp1),
       I would definitely need a UC to act as a temporary place holder.

       Also note that if you choose to go without *C quantities, then you need to remove
       the Reset step in Model::solve.
*/

		node[ce].dnsC    = dnsC    = node[ce].UC[0];
		node[ce].vxC     = vxC     = node[ce].UC[1]/dnsC;
		node[ce].vyC     = vyC     = node[ce].UC[2]/dnsC;
		node[ce].eneC    = eneC    = node[ce].UC[4]/dnsC;
		node[ce].tempC   = tempC   = (1.0/R)*(gam - 1.0) * ( eneC - 0.5*(vxC*vxC + vyC*vyC) );
		node[ce].enthalC = enthalC = R*gam*tempC/(gam - 1.0) + 0.5*(vxC*vxC + vyC*vyC);
		node[ce].prsC    = prsC    = R*dnsC*tempC;

		/* set F Corrector */                  /* set G Corrector */

		node[ce].FC[0] = dnsC*vxC;             node[ce].GC[0] = dnsC*vyC;
		node[ce].FC[1] = dnsC*vxC*vxC + prsC;  node[ce].GC[1] = dnsC*vyC*vxC;
		node[ce].FC[2] = dnsC*vxC*vyC;         node[ce].GC[2] = dnsC*vyC*vyC + prsC;
		node[ce].FC[3] = 0.0;                  node[ce].GC[3] = 0.0;
		node[ce].FC[4] = dnsC*vxC*enthalC;     node[ce].GC[4] = dnsC*vyC*enthalC;
	    }
	}

	return 1;

}

