/* This program initializes the flow field for a compressible
   CFD flow field.   The initialized variables are F, G, H, U which are
   the vectors holding the conserved quantities in the system of
   equations we are solving.  This is for both their predictor and
   corrector quantities as well.  This is the same notation as used on
   page 261 of:

Hirsch, C., Numerical Computation of Internal and External Flows Vol. 2: Computational
   Methods for Inviscid and Viscous Flows, John Wiley & Sons, Chichester, 1994.

                     Updated 2/2/06

    SLFCFD source file
    Version:  1.1
    Copyright (C) 2005, 2006, 2007, 2008  San Le

    The source code contained in this file is released under the
    terms of the GNU Library General Public License.

*/

#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "cfdconst.h"
#include "cfdstruct.h"

int numnp, sof, nx, ny, nz;
double dx, dy, dz;
int flag_2D, flag_3D;
double tolerance;

int InitCalc(COM_NODE *node)
{
	int i, j, k, node_num;
	double fdum;
	double dns, temp, prs, vx, vy, vz, enthal, ene;
/*
	double dnsA = 1.0, dnsB = 0.2008;
	double tempA = 1.0/1.4, tempB = 1.2/1.4;
	double vA = 1.0, vB = 0.0;
*/

	for (i = 0; i < numnp; i++)
	{
	   node[i].dnsC    = node[i].dnsP    = dns    = node[i].dns;
	   node[i].tempC   = node[i].tempP   = temp   = node[i].temp;
	   node[i].prsC    = node[i].prsP    = prs    = node[i].prs;
	   node[i].eneC    = node[i].eneP    = ene    = node[i].ene;
	   node[i].enthalC = node[i].enthalP = enthal = node[i].enthal;
	   node[i].vxC     = node[i].vxP     = vx     = node[i].vx;
	   node[i].vyC     = node[i].vyP     = vy     = node[i].vy;
	   node[i].vzC     = node[i].vzP     = vz     = node[i].vz;

	   if(flag_2D)
	   {
		node[i].vzC = node[i].vzP = node[i].vz = vz = 0.0;
	   }

/* For current quantities */

	   node[i].F[0] = dns*vx;            node[i].G[0] = dns*vy;            node[i].H[0] = dns*vz; 
	   node[i].F[1] = dns*vx*vx + prs;   node[i].G[1] = dns*vy*vx;         node[i].H[1] = dns*vz*vx; 
	   node[i].F[2] = dns*vx*vy;         node[i].G[2] = dns*vy*vy + prs;   node[i].H[2] = dns*vz*vy; 
	   node[i].F[3] = dns*vx*vz;         node[i].G[3] = dns*vy*vz;         node[i].H[3] = dns*vz*vz + prs; 
	                                                         if(flag_2D)   node[i].H[3] = 0.0;
	   node[i].F[4] = dns*vx*enthal;     node[i].G[4] = dns*vy*enthal;     node[i].H[4] = dns*vz*enthal; 

/* For predictor quantities */

	   node[i].FP[0] = dns*vx;           node[i].GP[0] = dns*vy;           node[i].HP[0] = dns*vz; 
	   node[i].FP[1] = dns*vx*vx + prs;  node[i].GP[1] = dns*vy*vx;        node[i].HP[1] = dns*vz*vx; 
	   node[i].FP[2] = dns*vx*vy;        node[i].GP[2] = dns*vy*vy + prs;  node[i].HP[2] = dns*vz*vy; 
	   node[i].FP[3] = dns*vx*vz;        node[i].GP[3] = dns*vy*vz;        node[i].HP[3] = dns*vz*vz + prs; 
	                                                          if(flag_2D)  node[i].HP[3] = 0.0;
	   node[i].FP[4] = dns*vx*enthal;    node[i].GP[4] = dns*vy*enthal;    node[i].HP[4] = dns*vz*enthal; 

/* For corrector quantities */

	   node[i].FC[0] = dns*vx;           node[i].GC[0] = dns*vy;           node[i].HC[0] = dns*vz; 
	   node[i].FC[1] = dns*vx*vx + prs;  node[i].GC[1] = dns*vy*vx;        node[i].HC[1] = dns*vz*vx; 
	   node[i].FC[2] = dns*vx*vy;        node[i].GC[2] = dns*vy*vy + prs;  node[i].HC[2] = dns*vz*vy; 
	   node[i].FC[3] = dns*vx*vz;        node[i].GC[3] = dns*vy*vz;        node[i].HC[3] = dns*vz*vz + prs; 
	                                                          if(flag_2D)  node[i].HC[3] = 0.0;
	   node[i].FC[4] = dns*vx*enthal;    node[i].GC[4] = dns*vy*enthal;    node[i].HC[4] = dns*vz*enthal; 

/* For U current/predictor/corrector quantities */

	   node[i].U[0] = dns;      node[i].UP[0] = dns;      node[i].UC[0] = dns; 
	   node[i].U[1] = dns*vx;   node[i].UP[1] = dns*vx;   node[i].UC[1] = dns*vx; 
	   node[i].U[2] = dns*vy;   node[i].UP[2] = dns*vy;   node[i].UC[2] = dns*vy; 
	   node[i].U[3] = dns*vz;   node[i].UP[3] = dns*vz;   node[i].UC[3] = dns*vz; 
	   node[i].U[4] = dns*ene;  node[i].UP[4] = dns*ene;  node[i].UC[4] = dns*ene; 

	   /*printf(" %4d %16.8e %16.8e %16.8e %16.8e %16.8e %16.8e %16.8e\n",
	       i, temp, vx, vy, vz, prs, ene, dns);*/ 
	   /*printf(" %4d %4d %4d %4d %12.6e %12.6e %12.6e\n",
	       i, j, k, number, no->getF(1), no->getG(1), no->getH(1));*/
	}

	return 1;
}

