/* This program draws the streaklines for the program which uses the
   finite difference method to calculate a CFD flow field.

                     Updated 4/8/08

    SLFCFD source file
    Version:  1.1
    Copyright (C) 2006, 2007  San Le

    The source code contained in this file is released under the
    terms of the GNU Library General Public License.

*/

#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "cfdconst.h"
#include "cfdstruct.h"
#include "../common/shape_struct.h"

#if COMPRESSIBLE
#define NODE    COM_NODE
#endif
#if INCOMPRESSIBLE
#define NODE    INC_NODE
#endif
#if GUI
#define NODE    GR_NODE
#endif


int cfdshape(SHAPE *, double , double );

extern double dx, dy, dz;
extern double dt, time_current;
extern int nstep, cstep;
extern int imax, jmax;


int streaklines(int current_step, FLOWLINES streak, NODE *node)
{

	int i, i2, j, k, k2, dum, dum2;
	int ce, xm1, xm2, xp1, xp2, ym1, ym2, yp1, yp2, xm1ym1, xm1yp1, xp1ym1, xp1yp1;
	int index_i, index_j, num_pt0;
	double perx, pery, xi, xi2, eta, eta2, fdum, fdum2;
	SHAPE sh;

	fdum2 = 1.0;
	num_pt0 = streak.num_pt0;
/* The variables below generate the streaklines
*/
	for( k = 0; k < num_pt0; ++k )
	{

/* Initialize the next set of streaklines which start at streak.pt0. */

	    streak.pt[num_pt0*current_step+k].x = streak.pt0[k].x;
	    streak.pt[num_pt0*current_step+k].y = streak.pt0[k].y;

/* Continue updating the streaklines from before. */

	    for( k2 = 0; k2 < current_step; ++k2 )
	    {
		i = (int)(streak.pt[num_pt0*k2+k].x/dx);
		j = (int)(streak.pt[num_pt0*k2+k].y/dy);

		ce = jmax*i + j;
		xm1 = jmax*(i-1) + j;
		xp1 = jmax*(i+1) + j;
		ym1 = jmax*i + j-1;
		yp1 = jmax*i + j+1;
		xp1yp1 = jmax*(i+1) + j+1;
		xm1yp1 = jmax*(i-1) + j+1;
		xp1ym1 = jmax*(i+1) + j-1;

		if( i < imax-1 && i > 0 && j < jmax-1 && j > 0 ) 
		{
		    perx = streak.pt[num_pt0*k2+k].x/dx - (double)(i);
		    pery = streak.pt[num_pt0*k2+k].y/dy - (double)(j);

		    xi  = 2.0*perx - 1.0;
		    eta = 2.0*pery - 1.0;

/*     Because I am using a staggared grid for u(i,j) and v(i,j) which do not fall on node
       points, xi and eta have to be determined based on this shift.

       First, -1 <= xi <= 1.0, and xi = 0.0 is where the node of u lies.  This means that I check whether
       at xi > 0.0. in which case I interpolate between node[ce].vx and node[xp1].vx.  If not, then the interpolation
       is between node[xm1].vx and node[ce].vx. 

       In addition, once it is determined which side of 0.0 xi lies on, I will continue using xi for
       interpolation purposes for u.  The only consideration I have to make is either I do xi + 1.0 if
       xi < 0.0 (because I'm interpolating from node[xm1].vx to node[ce].vx or xi - 1.0 if xi > 0.0 because I'm
       interpolating from node[ce].vx to node[xp1].vx. 

       The same is done for v with respect to eta.

       On October 9th, 2007, I rechecked the algorithm below and it seemed to still be good.
*/
		    if( xi > 0.0 )
		    {
			xi2 = xi - 1.0;
			cfdshape( &sh, xi2, eta);

			fdum = sh.Nhat[0].dx0*node[ce].vx +
			  sh.Nhat[1].dx0*node[xp1].vx +
			  sh.Nhat[2].dx0*node[xp1yp1].vx +
			  sh.Nhat[3].dx0*node[yp1].vx;
		    }
		    else
		    {
			xi2 = xi + 1.0;
			cfdshape( &sh, xi2, eta);

			fdum = sh.Nhat[0].dx0*node[xm1].vx +
			  sh.Nhat[1].dx0*node[ce].vx +
			  sh.Nhat[2].dx0*node[yp1].vx +
			  sh.Nhat[3].dx0*node[xm1yp1].vx;
		    }

		    streak.pt[num_pt0*k2+k].x =
			streak.pt[num_pt0*k2+k].x + fdum2*dt*fdum;

		    if( eta > 0.0 )
		    {
			eta2 = eta - 1.0;
			cfdshape( &sh, xi, eta2);

			fdum = sh.Nhat[0].dx0*node[ce].vy +
			  sh.Nhat[1].dx0*node[xp1].vy +
			  sh.Nhat[2].dx0*node[xp1yp1].vy +
			  sh.Nhat[3].dx0*node[yp1].vy;
		    }
		    else
		    {
			eta2 = eta + 1.0;
			cfdshape( &sh, xi, eta2);

			fdum = sh.Nhat[0].dx0*node[ym1].vy +
			  sh.Nhat[1].dx0*node[xp1ym1].vy +
			  sh.Nhat[2].dx0*node[xp1].vy +
			  sh.Nhat[3].dx0*node[ce].vy;
		    }

		    streak.pt[num_pt0*k2+k].y =
			streak.pt[num_pt0*k2+k].y + fdum2*dt*fdum;

		    /*streak.pt[num_pt0*k2+k].x =
			streak.pt[num_pt0*k2+k].x + fdum2*dt*(perx*(node[xp1].vx - node[ce].vx) + node[ce].vx);
		    streak.pt[num_pt0*k2+k].y =
			streak.pt[num_pt0*k2+k].y + fdum2*dt*(perx*(node[yp1].vy - node[ce].vy) + node[ce].vy);*/
		} 
		else
		{
		    streak.pt[num_pt0*k2+k].x = -100000.0;
		    streak.pt[num_pt0*k2+k].y = -100000.0;
		}
	    }
	}

	return 1;

}


