/* This program uses the finite difference method to calculate an incompressible
   laminar CFD flow field.  It reads in the data and calculates the various
   quantities for each point in the rectangular CFD mesh.

   The algorithm for the code is based on the QUICKEST SCHEME.
   The SOR method is used to calculate the pressure corrector.
   The main references are:

Davis, R. W., and E. F. Moore, Journal of Fluid Mechanics, 116, 475 (1982).

Davis, R. W., E. F. Moore, and L. P. Purtell, Physics of Fluids, 27, 46 (1984).

   More is given on this method on page 347 of:

Fletcher, Clive A. J., Computational Techniques for Fluid Dynamics: Vol II, second
        edition, John Wiley & Sons Inc., New York, 1991.

                     Updated 1/7/10

    SLFCFD source file
    Version:  1.1
    Copyright (C) 2006-2010  San Le

    The source code contained in this file is released under the
    terms of the GNU Library General Public License.

*/

#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <time.h>
#include "../compressible/cfdconst.h"
#include "../compressible/cfdstruct.h"
#include "../AMR/amr.h"

int writer_2D( BOUND , MATL *, char *, INC_NODE * );

int GnuplotWriter( BODY *, char *, FLOWLINES, FLOWLINES );

int pathlines(int , FLOWLINES ,  INC_NODE * );

int streaklines(int , FLOWLINES ,  INC_NODE * );

int reset( INC_NODE *, SUBGRID );

int reset0( BOUND , INC_NODE *, SUBGRID);

int applyBC_free( BOUND bc, INC_NODE *, SUBGRID );

int Cyclic_Reduction(double *, double *, double *, double *, int , int );

int refineGridSet( double *, double *, SUBGRID , SUBGRID );

int coarseGridSet_AMR( double *, double *, double *, double *, double *, double *,
	SUBGRID , SUBGRID );

int PoisConjGrad_AMR( double *, double *, double *, SUBGRID );

int refineGridSetBC( double *, double *, double *, double *, double *, double *,
	SUBGRID , SUBGRID );

int refineGridSet_AMR( double *, double *, double *, double *, double *, double *,
	SUBGRID , SUBGRID );

int setSubgrid( SUBGRID * );

int PoisConjGradCoarse( double *, double *, double *, SUBGRID, SUBGRID );

int PoisConjGrad( double *, double *, double *, SUBGRID );

int MassFlow( double *, INC_NODE *, SUBGRID, double * );

int sor( BOUND , INC_NODE *, SUBGRID ); 

int face_velocities( BOUND, INC_NODE *, SUBGRID );

int quickest( BOUND , INC_NODE *, SUBGRID );

int applyBC_freestream( BOUND , INC_NODE *, SUBGRID );

int applyBC_free2( BOUND , INC_NODE *, SUBGRID );

int applyBC_noslip( BOUND , INC_NODE *, SUBGRID );

int applyBC_inflow( BOUND , INC_NODE *, SUBGRID );

int create_body( BODY *);

int bcset_noslip( BOUND *, BODY *, int *, INC_NODE * );

int bcset( BOUND *, int *, INC_NODE * );

int qikreader( MATL *, char *, INC_NODE *, FILE *, FLOWLINES, FLOWLINES );

int cfdMemory( double **, int , int **, int , MATL **, int , XYZF **, int , XYZI **, int ,
	XYZI2 **, int , XY_ZX_YZ **, int , INC_NODE **, int, SUBGRID **, int );

int Flowlines_File( FLOWLINES * );


int dof, nmat, numel, numnp, sof, nx, ny, nz, num_bodies;
double dx, dy, dz;
int flag_2D = 1, flag_3D = 0, print_flag = 0, debug_flag = 0, reverse_flag = 0;
double dt, time_current, err = 0.0;         /* time step for time integration */
int vmag_read_flag = 0;
int nstep, cstep, ANGLE;
int imax, jmax;
double q, alp, Re;
double umax;
int umax_i, umax_j;


int main(int argc, char **argv)
{
	int i, i2, j, j2, k, dum, dum2, node_num, node_num2, bod_num, pt;
	int check, name_length, counter, MemoryCounter, file_loc;
	int ce, ce2, xm1, xm2, xp1, xp2, xp3, ym1, ym2, yp1, yp2, xm1ym1, xm1yp1, xp1ym1, xp1yp1;
	double time, time_final, length, width;
	double uxm1, vym1, pxm1, pxp1, pym1, pyp1;
	INC_NODE *node, *node1, *node2, *node3, *mem_INC_NODE;
	SUBGRID subgrid0, *subgrid1, *subgrid2, *subgrid3, *mem_SUBGRID;
	XYZI *mem_XYZI;
	XYZI2 *mem_XYZI2; 
	XYZF *mem_XYZF;
	XY_ZX_YZ *mem_XY_ZX_YZ;
	int *mem_int, sofmi, sofmf, sofmXYZF, sofmXYZI, sofmXYZI2, sofmXY_ZX_YZ,
		sofmINC_NODE, sofmSUBGRID, ptr_inc;
	MATL *matl; 
	double fdum, fdum2;
	double dns, temp, vx, vy, pres, ene, enthal, vmag;
	double *mem_double;
	double *mass_flow;
	double usq1, usq2, vsq1, vsq2, uv1, uvQ;
	double px, py; 
	char name[30], object_name[30], buf[ BUFSIZ ], *ccheck, one_char;
	FILE *o1;
	BOUND bc;
	BODY *body;
	FLOWLINES streak, path;
	int read_data_flag;
	long timec;
	long timef;

	sof = sizeof(double);

#if 0
	printf("Is there an input file?\n");
	printf("0)No\n");
	printf("1)Yes\n");
	scanf( "%d",&read_data_flag);
#endif

	printf("What is the name of the file containing the \n");
	printf("incompressible finite difference CFD data? (example: a0)\n");
	scanf( "%30s",name);

/*   o1 contains all the structural data
     streak.file_in is the initial points of the streaklines
     streak.file_out is the initial points of the streaklines
     path.file_in is the initial points of the streaklines
     path.file_out is the initial points of the streaklines
*/

	o1 = fopen( name,"r" );
	if(o1 == NULL ) {
		printf("Can't find file %30s\n", name);
		exit(1);
	}

	fgets( buf, BUFSIZ, o1 );
	fscanf( o1, "%d %d %d", &numnp, &nx, &ny);
	imax = nx;
	jmax = ny;
	nmat = 1;

/* Check if there is additional coordinate z data and that this is a 3-D file.  */

	flag_3D = 0;
	flag_2D = 1;
	nz = 1;
	while(( one_char = (unsigned char) fgetc(o1)) != '\n')
	{
		if(one_char != ' ' )
		{
		    ungetc( one_char, o1);
		    fscanf( o1,"%d", &nz);
		    flag_3D = 1;
		    flag_2D = 0;
		    break;
		}
	}
	numnp = nx*ny*nz;
	dof=numnp;
	fscanf( o1,"\n");

	name_length = strlen(name);
	if( name_length > 25) name_length = 25;

/* Set the name of the input streakline file */

	memset(streak.name_in,0,30*sizeof(char));

	streak.read_flag = 0;
	ccheck = strncpy(streak.name_in, name, name_length);
	if(!ccheck) printf( " Problems with strncpy \n");

	ccheck = strncpy(streak.name_in+name_length, ".streak", 7);
	if(!ccheck) printf( " Problems with strncpy \n");

/* Check for the existence of the streakline input file as well as its properties */

	check = Flowlines_File( &streak );
	if(!ccheck) printf( " Problems with Flowlines_File \n");

/* Set the name of the input pathline file */

	memset(path.name_in,0,30*sizeof(char));

	ccheck = strncpy(path.name_in, name, name_length);
	if(!ccheck) printf( " Problems with strncpy \n");

	ccheck = strncpy(path.name_in+name_length, ".path", 5);
	if(!ccheck) printf( " Problems with strncpy \n");

/* Check for the existence of the pathline input file as well as its properties */

	path.read_flag = 0;
	check = Flowlines_File( &path );
	if(!ccheck) printf( " Problems with Flowlines_File \n");

	print_flag = 0;
	printf("%d\n", argc);
	if(argc > 2)
	{
		printf("usage %s -p\n", name);
		printf("Too many arguments.\n");
		exit(1);
	}
	if(argc == 2 )
	{
	    if( !strncmp(argv[1], "-p", 2))
	    {
		print_flag = 1;
	    }
	    else
	    {
		printf("usage %s -p only\n",name);
		exit(1);
	    }
	}

	if(flag_3D) printf( "total nodes:%d   nodes in x:%d nodes in y:%d nodes in z:%d\n", numnp, nx, ny, nz);
	else        printf( "total nodes:%d   nodes in x:%d nodes in y:%d\n", numnp, nx, ny);
	fgets( buf, BUFSIZ, o1 );
	printf( "\n");

	fscanf( o1, "%lf %d %d %lf %lf\n", &dt, &nstep, &cstep, &time_current, &err);
	if( nstep < cstep)
	{
	    printf("Number of steps must be equal or greater than current step.\n");
	    exit(1);
	}
	if( nstep == 0)
	{
	    printf("Number of steps cannot be zero.\n");
	    exit(1);
	}

	printf( "Time step:     %16.8e\n", dt);
	printf( "number of steps:  %4d\n", nstep);
	printf( "current step:     %4d\n", cstep);
	printf( "current time:  %16.8e\n", time_current);
	printf( "SOR tolerence: %16.8e\n", err);

	printf( "\n");

	fgets( buf, BUFSIZ, o1 );
	printf( "\n");

	streak.nstep_out = nstep;
	path.nstep_out = nstep;

/*   Begin allocation of meomory */

	MemoryCounter = 0;

/* For the doubles */
	sofmf = numnp;
	MemoryCounter += sofmf*sizeof(double);
	printf( "\n Memory requrement for doubles is %15d bytes\n",MemoryCounter);

/* For the integers */
	sofmi= 2 + 4*numnp;
	MemoryCounter += sofmi*sizeof(int);
	printf( "\n Memory requrement for integers is %15d bytes\n",MemoryCounter);

/* Note that I allocate 2*numnp for the nodal boundaries rather than just numnp
   because some nodes may have more than 1 boundary condition, specifically the
   no-slip corner boundary which results in a node being no-slip in more than 1
   direction.  So I can have, for instance:

      bc[0].noslp.xm[3] = 5;
      bc[0].noslp.ym[3] = 5;

   This potential for having more than one boundary condition is why I have allocated
   2*numnp.
*/

/* For the XYZF doubles */
	sofmXYZF = streak.num_pt0 + streak.num_pt0*streak.nstep_out + 
		path.num_pt0 + path.num_pt0*path.nstep_out;
	MemoryCounter += sofmXYZF*sizeof(XYZF);
	printf( "\n Memory requrement for XYZF doubles is %15d bytes\n",MemoryCounter);

/* For the XYZI integers */
	sofmXYZI = 1;
	MemoryCounter += sofmXYZI*sizeof(XYZI);
	printf( "\n Memory requrement for XYZI integers is %15d bytes\n",MemoryCounter);

/* For the XYZI2 integers */
	sofmXYZI2=7;
	MemoryCounter += sofmXYZI2*sizeof(XYZI2);
	printf( "\n Memory requrement for XYZI2 integers is %15d bytes\n",MemoryCounter);

/* For the XY_ZX_YZ integers */
	sofmXY_ZX_YZ=1;
	MemoryCounter += sofmXY_ZX_YZ*sizeof(XY_ZX_YZ);
	printf( "\n Memory requrement for XY_ZX_YZ integers is %15d bytes\n",MemoryCounter);

/* For the INC_NODE structs */
	sofmINC_NODE=2*numnp + 1;
	MemoryCounter += sofmINC_NODE*sizeof(INC_NODE);
	printf( "\n Memory requrement for INC_NODE integers is %15d bytes\n",MemoryCounter);

/* For the SUBGRID structs */
	sofmSUBGRID=6*numnp;
	MemoryCounter += sofmSUBGRID*sizeof(SUBGRID);
	printf( "\n Memory requrement for SUBGRID integers is %15d bytes\n",MemoryCounter);

	check = cfdMemory( &mem_double, sofmf, &mem_int, sofmi, &matl, nmat, &mem_XYZF, sofmXYZF,
		&mem_XYZI, sofmXYZI, &mem_XYZI2, sofmXYZI2, &mem_XY_ZX_YZ, sofmXY_ZX_YZ, &mem_INC_NODE,
		sofmINC_NODE, &mem_SUBGRID, sofmSUBGRID );
	if(!check) printf( " Problems with cfdMemory\n");

/* For the doubles */
	                                      ptr_inc=0;
	mass_flow =(mem_double+ptr_inc);      ptr_inc += numnp;

/* For the INC_NODE structs*/
	                                      ptr_inc = 0;
	node=(mem_INC_NODE+ptr_inc);          ptr_inc += numnp;
	node1=(mem_INC_NODE+ptr_inc);         ptr_inc += numnp;

/* For the integers */
	                                      ptr_inc = 0;
	bc.num_none=(mem_int+ptr_inc);        ptr_inc += 1;
	bc.num_prs0=(mem_int+ptr_inc);        ptr_inc += 1;

/* Note that bcset below has to be corrected for mem_int so that it has the correct
   offset for mem_int, i.e.:

	check = bcset( &bc, (mem_int+2), node);

	check = bcset_noslip( &bc, body, (mem_int+2+2*numnp), node);
*/

/* For the XYZF doubles */
	                                      ptr_inc = 0;
	streak.pt0=(mem_XYZF+ptr_inc);        ptr_inc += streak.num_pt0;
	streak.pt=(mem_XYZF+ptr_inc);         ptr_inc += streak.num_pt0*streak.nstep_out;
	path.pt0=(mem_XYZF+ptr_inc);          ptr_inc += path.num_pt0;
	path.pt=(mem_XYZF+ptr_inc);           ptr_inc += path.num_pt0*path.nstep_out;

/* For the XYZI integers */
	                                      ptr_inc = 0;
	bc.num_vel0=(mem_XYZI+ptr_inc);       ptr_inc += 1;

/* For the XYZI2 integers */
	                                      ptr_inc = 0;
	bc.num_symm=(mem_XYZI2+ptr_inc);      ptr_inc += 1;
	bc.num_far=(mem_XYZI2+ptr_inc);       ptr_inc += 1;
	bc.num_free=(mem_XYZI2+ptr_inc);      ptr_inc += 1;
	bc.num_inflow=(mem_XYZI2+ptr_inc);    ptr_inc += 1;
	bc.num_frestr=(mem_XYZI2+ptr_inc);    ptr_inc += 1;
	bc.num_free2=(mem_XYZI2+ptr_inc);     ptr_inc += 1;
	bc.num_noslp=(mem_XYZI2+ptr_inc);     ptr_inc += 1;

/* For the XY_ZX_YZ integers */
	                                          ptr_inc = 0;
	bc.num_noslp_prs=(mem_XY_ZX_YZ+ptr_inc);  ptr_inc += 1;

/* For the SUBGRID structs*/
	                                      ptr_inc = 0;
	subgrid1=(mem_SUBGRID+ptr_inc);       ptr_inc += 2*numnp;
	subgrid2=(mem_SUBGRID+ptr_inc);       ptr_inc += 2*numnp;
	subgrid3=(mem_SUBGRID+ptr_inc);       ptr_inc += 2*numnp;

	for( i = 0; i < numnp; ++i )
	{
	     node[i].bc_type = 0;
	}

	for( i = 0; i < 2*numnp; ++i )
	{
		subgrid1[i].refine_num = 1;
		subgrid2[i].refine_num = 2;
		subgrid3[i].refine_num = 3;
	}

/* Initialzie pressure */

	for( i = 0; i < imax; ++i )
	{
		for( j = 0; j < jmax; ++j )
		{
		   ce  = jmax*i + j;
		   node[ce].prs = 0.0;
		}
	}


	timec = clock();
	timef = 0;

	check = qikreader( matl, name, node, o1, path, streak );
	if(!check) printf( " Problems with qikreader\n");

	for( i = 0; i < nx; ++i )
	{
	    for( j = 0; j < ny; ++j )
	    {
		ce = ny*i + j;
		node[ce].vxQ = node[ce].vx;
		node[ce].vyQ = node[ce].vy;
		/*printf(  "llll %16.8e %16.8e %16.8e %4d\n", node[ce].vx, node[ce].vy, node[ce].prs, ce);*/
	    }
	}

/* Establishing initial conditions for u,v, and pressure */

	Re = matl[0].Re;
	matl[0].gam = 1.0;

	q = 0.375;
	alp = 1.0/Re;

/* Set the boundary condition variable, bc, as well as allocate memory for each individual
   boundary condition.  
*/

	check = bcset( &bc, (mem_int+2), node);
	if(!check) printf( " Problems with bcset\n");

	for( i = 0; i < streak.num_pt0; ++i )
	{
	    if(print_flag) printf(  "streak %4d   %16.8e %16.8e\n", i, streak.pt0[i].x, streak.pt0[i].y);
	}

	for( i = 0; i < path.num_pt0; ++i )
	{
/* Set the initial pathline values. */

	    if(print_flag) printf(  "path %4d   %16.8e %16.8e\n", i, path.pt0[i].x, path.pt0[i].y);
	    path.pt[i].x = path.pt0[i].x;
	    path.pt[i].y = path.pt0[i].y;

	}

/* Allocate memory for the bodies in the flow.  */

	if(num_bodies)
	{
	    body = (BODY *)calloc(num_bodies,sizeof(BODY));
	    if(!body )
	    {
		printf( "failed to allocate memory for body structures\n ");
		exit(1);
	    }
	    dum = 2*bc.num_noslp[0].xm; /* twice the amount necessary, just in case. */
	    for( bod_num = 0; bod_num < num_bodies; ++bod_num )
	    {
		body[bod_num].nd = (J_I1_I2_INDX *)calloc(dum,sizeof(J_I1_I2_INDX));
		body[bod_num].coord = (double *)calloc(2*dum*nsd,sizeof(double));
	    }
	}

	check = bcset_noslip( &bc, body, (mem_int+2+2*numnp), node);
	if(!check) printf( " Problems with bcset_noslip\n");

/* This is only used for generating points which draw the body.  It is not needed for calculation. */

	if(num_bodies)
	{
	    check = create_body( body);
	    if(!check) printf( " Problems with create_body\n");
	}

/* I define subgrid0 to be the original coarsest mesh. */

	subgrid0.jstart = 0;
	subgrid0.jend = jmax-1;
	subgrid0.istart = 0;
	subgrid0.iend = imax-1;
	subgrid0.dx = dx;
	subgrid0.dy = dy;
	subgrid0.dt = dt;
	subgrid0.nx = nx;
	subgrid0.ny = ny;
	subgrid0.refine_num = 0;

/* reverse_flag was used to test the symmetry of the code.  Most of the example files have
   flows running from left to right.  To see if the code would produce the same results with
   the fluid running from bottom to top, I use the reverse flag which will transpose the
   quantities in the x and y directions.  This made the results the same as when they were
   running from left to right.  The 2 input files I used were:

           a-15    flow from left to right with a -15 degree angle of attack
           ay-15   flow from bottom to top with a 45 degree angle of attack

   The test was successful, so I am setting:

           reverse_flag = 0  

   in the 3rd line below.
*/

	if ( node[0].vx > 0.5) reverse_flag = 0;
	if ( node[0].vy > 0.5) reverse_flag = 1;

	reverse_flag = 0;

	/*check = quickest( bc, node, subgrid0);
	check = reset( node, subgrid0);*/

/* Begin calculation of flow over time */

	time_final = ((double)nstep)*dt;

	time = ((double)cstep)*dt;
	counter = cstep;

	while (time < time_final - dt/2.0 )
	{

		umax = 0.0;

		check = face_velocities( bc, node, subgrid0);

/* The no slip is applied for the face velocities */

		check = quickest( bc, node, subgrid0);

		/*check = MarkerAndCell( bc, node, subgrid0);*/

/* The no slip is applied */

		check = applyBC_noslip( bc, node, subgrid0);

		if(umax > 5.0)
		{
		    for( i = 0; i < imax; ++i )
		    {
			for( j = 0; j < jmax; ++j )
			{
			   ce  = jmax*i + j;
			   node[ce].prs = 0.0;
			}
		    }
		}

		if(umax > 10.0)
		{
		    printf(  "The velocity has become too large: %16.8e %16.8e %16.8e %5d\n",
			umax, time, dt, counter);
		    time = time_final;
		}

/* Below is the bondary condition for the outflow at the right. 
   It represents linearally extrapolating the velocity and pressure
   from the previous 2 profiles.
*/
		check = applyBC_free2( bc, node, subgrid0);

/* Below is the bondary condition for the upper and lower part of the
   grid.  It represents saying the velocity is the freestream
   velocity at the top and bottom of the grid, i.e. dv/dn and du/dn
   are zero where n is the normal to the top of the grid.
*/

		check = applyBC_freestream( bc, node, subgrid0);

/*     The no slip is reapplied */

		check = applyBC_noslip( bc, node, subgrid0);

/* This concludes the creation of the velocity field whithout the
   pressure correction.  sor below will then determine the pressure
   field.  The equation for this is based on Equation 17.13 page 339 in

     Fletcher, Clive A. J., Computational Techniques for Fluid Dynamics: Vol II, 2nd
	Ed., Springer-Verlag, Berlin, 1991.

   where the F is replaced with vxQ and G is vyQ.
*/

/*
		subgrid1[0].dt = dt;
		subgrid1[0].dx = dx;
		subgrid1[0].dy = dy;
*/


/* Set pressure back to zero at the last step */

		if( time > time_final - dt )
		{
		    for( i = 0; i < imax; ++i )
		    {
			for( j = 0; j < jmax; ++j )
			{
			   ce  = jmax*i + j;
			   node[ce].prs = 0.0;
			}
		    }
		}


/* Use the temporary node velocities at this point to store the velocities
   at the farfield and freestream boundary conditions.  These are used in
   the sor function and improve stability.
*/

		check = reset0( bc, node, subgrid0);

		check = sor( bc, node, subgrid0); 

/* Below is the bondary condition for the inflow at the left.  It represents
   linearally extrapolating the pressure from the 2 nodes to the right.  I
   decided not to use it since this extrapolation of pressure is already done
   in "sor".
*/
		 /*check = applyBC_inflow( bc, node, subgrid0);*/

/* I considered setting all pressures internal to the body to zero after every step. 
   This did not improve the results.
*/

		/*for( j = ypbox + 1; j < ypbox + hbox; ++j )
		{
			for( i = xpbox + 1; i < xpbox + lbox; ++i )
			{
			   ce  = ny*i + j;
			   node[ce].prs = 0.0;
			}
		}*/

/*     The no slip is applied again. */

		check = applyBC_noslip( bc, node, subgrid0);

/* The pressure field has been calculated from sor and now the pressure correction
   can be made below.
*/

#if 1
		for( i = 1; i < imax-1; ++i )
		{
		    for( j = 1; j < jmax-1; ++j )
		    {
			ce  = jmax*i + j;
			xp1 = jmax*(i+1) + j;
			yp1 = jmax*i + (j+1);

			node[ce].vxQ = node[ce].vxQ - dt/dx*(node[xp1].prs - node[ce].prs);
			node[ce].vyQ = node[ce].vyQ - dt/dy*(node[yp1].prs - node[ce].prs);
			/*printf("%4d %4d %4d %16.8e  %16.8e  %16.8e\n",
			    i, j, ce, node[xp1].prs, node[ce].prs, node[yp1].prs);*/
		    }
		}
#endif

		check = MassFlow( mass_flow, node, subgrid0, &fdum);
		if(!check) printf( " Problems with MassFlow\n");

/* Now u and v are set to equal uQ and vQ */

		check = reset( node, subgrid0);

		if(flag_2D)
		{
/* Update the value of all the streaklines points. */

		    if(streak.read_flag) check = streaklines( counter, streak, node);

/* Calculate the next point of the pathlines. */

		    if(path.read_flag) if(counter < nstep - 1 )
			check = pathlines( counter, path, node);
		}

		time += dt;
		++counter;
		printf( " Time: %16.8e  step: %5d\n", time, counter);

	}

/* I calculate cstep and time_current differently than in the compressible code because
   sometimes the calculation has to stop if the velocity or pressure become too high.
*/

	cstep = counter;
	time_current = dt*(double)(counter);

/* Write out the streakline and obstruction data to a file which can be read
   into "gnuplot".
*/

	check = GnuplotWriter( body, name, path, streak);
	if(!check) printf( " Problems with GnuplotWriter\n");

/*here:       printf(  " had to stop %i %i  \n", i,k );*/


	if(!reverse_flag) 
	{
	    for( i = 0; i < nx; ++i )
	    {
		for( j = 0; j < ny; ++j )
		{
		    node_num = ny*i + j;

		    node[node_num].i = i;
		    node[node_num].j = j;

		    dns    = 0.0;
		    temp   = 0.0;
		    vx     = node[node_num].vx;
		    vy     = node[node_num].vy;
		    pres   = node[node_num].prs;
		    ene    = 0.0;
		    enthal = 0.0;

		    fdum = (vx*vx + vy*vy);
		    node[node_num].vmag = sqrt(fdum);
		}
	    }

#if 0
	    memset(name,0,30*sizeof(char));
	    ccheck = strncpy(name, "abcd", 4);
	    if(!ccheck) printf( " Problems with strncpy \n");
#endif

	    check = writer_2D(bc, matl, name, node);
	    if(!check) printf( " Problems with writer_2D\n");
	}

	if(reverse_flag) 
	{
	    for( i = 0; i < nx; ++i )
	    {
		for( j = 0; j < ny; ++j )
		{
		    node_num = ny*i + j;
		    node_num2 = nx*j + i;

		    node1[node_num2].j = i;
		    node1[node_num2].i = j;

		    vx = node1[node_num2].vy    = node[node_num].vx;
		    vy = node1[node_num2].vx    = node[node_num].vy;
		    node1[node_num2].prs        = node[node_num].prs;
		    node1[node_num2].bc_type    = node[node_num].bc_type;
		    node1[node_num2].nsid       = node[node_num].nsid;

		    fdum = (vx*vx + vy*vy);
		    vmag = sqrt(fdum);
		    node[node_num].vmag         = vmag;
		    node1[node_num2].vmag       = node[node_num].vmag;
		}
	    }

#if 0
	    memset(name,0,30*sizeof(char));
	    ccheck = strncpy(name, "abcd", 4);
	    if(!ccheck) printf( " Problems with strncpy \n");
#endif

	    dum = nx;
	    dum2 = ny;
	    nx = dum2;
	    ny = dum;

	    check = writer_2D(bc, matl, name, node1);
	    if(!check) printf( " Problems with writer_2D\n");

	    nx = dum;
	    ny = dum2;
	}

	printf("%d ", j);

	timec = clock();
	printf("\n\n elapsed CPU = %lf\n\n",( (double)timec)/800.);

	free(matl);
	free(mem_double);
	free(mem_int);
	free(mem_XYZF);
	free(mem_XYZI);
	free(mem_XYZI2);
	free(mem_XY_ZX_YZ);
	free(node);
	free(mem_SUBGRID);
	
	return 1;
}


