/* This program creates the standard input data file for the 
   finite difference compressible/incompressible CFD flow field.  It
   reads from the file "meshtemp.input" and generates a rectangular
   mesh with the most likely boundary conditions on the perimeter.

                     Updated 1/15/09

    SLFCFD source file
    Version:  1.1
    Copyright (C) 2005-2009  San Le

    The source code contained in this file is released under the
    terms of the GNU Library General Public License.

*/

#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "../compressible/cfdconst.h"
#include "../compressible/cfdstruct.h"

#define SMALL   1.0e-20
#define pi      3.141592654

int sof, nx, ny, nz, num_bodies;
int flag_2D, flag_3D, reverse_flag = 0;
double dx, dy, dz, Re;
double dt, time_current;
int nstep, cstep;

int meshBCset( int *, char *);

int main (BOUND bc, MATL *matl, char *name, INC_NODE *node)
{
	int i, j, k, dum, node_num, numnp, bc_type, check, name_length, choice;
	char *ccheck, buf[ BUFSIZ ];
	int dum2,dum3,dum4;
	double fdum, fdum1, fdum2, fdum3;
	double dns, temp, vx, vy, prs, ene, enthal, vmag, mach;
	double gam, R, angle;
	char text[10], out[30];
	FILE *o1, *o3;

	o1 = fopen( "meshtemp.input","r" );
	if(o1 == NULL ) {
		printf("Can't find file meshtemp.input.\n"); 
		exit(1);
	} 

	o3 = fopen( "meshtemp.out","w" );

	fgets( buf, BUFSIZ, o1 );
	fscanf( o1, "%d\n", &choice);

	fgets( buf, BUFSIZ, o1 );
	fscanf( o1, "%d %d\n", &nx, &ny);
	fgets( buf, BUFSIZ, o1 );
	fscanf( o1, "%lf %d\n", &dt, &nstep);
	fgets( buf, BUFSIZ, o1 );
	fscanf( o1, "%lf %lf\n", &dx, &dy);
	fgets( buf, BUFSIZ, o1 );

	if(choice == 1)
	{
	    gam = 1.4;
	    fscanf( o1, "%lf %lf\n", &R, &gam);
	}
	else
	    fscanf( o1, "%lf\n", &Re);

	fgets( buf, BUFSIZ, o1 );
	fscanf( o1, "%lf\n", &angle);

	numnp = nx*ny;

	fprintf( o3, "total nodes, nodes in x, nodes in y for a typical input file\n");
	fprintf( o3, "%4d    %4d %4d\n", numnp, nx, ny);
	fprintf( o3, "Time step, number of steps, current step, current time");
	if(choice == 1) fprintf( o3, "\n");
	else fprintf( o3, ", SOR tolerence\n");
	fprintf( o3, "%16.8e %7d %7d %16.8e", dt, nstep, cstep, time_current);
	if(choice == 1) fprintf( o3, "\n");
	else fprintf( o3, "  %5.1e\n", 1.0e-4);
	fprintf( o3, "spatial discretization dx, dy\n");
	fprintf( o3, "%16.8e %16.8e\n", dx, dy);
	if(choice == 1)
	{
	    fprintf( o3, "Ideal Gas Constant, Ratio of specific heats\n");
	    fprintf( o3, "%16.8e %16.8e\n", R, gam);
	    fprintf( o3, "   i    j    bc           density      temperature  pressure     velocity x");
	    fprintf( o3, "   velocity y   enthalpy     energy       mach number\n");
	}
	else
	{
	    fprintf( o3, "Reynolds Number\n");
	    fprintf( o3, "%16.8e\n", Re);
	    fprintf( o3, "   i    j    bc           pressure     velocity x");
	    fprintf( o3, "   velocity y   velocity magnitude\n");

	}


	dum= 0;

	angle *= pi/180.0;

	vx      = 1.0*cos(angle);
	vy      = 1.0*sin(angle);

	if(choice == 1)
	{
	    prs = 1.7211e-01; dns = 2.0080e-01; temp = 8.5714e-01;
	    ene = 2.1429e+00; enthal = 3.0000e+00;

	    fdum = (vx*vx + vy*vy)/(R*gam*temp + SMALL);
	    mach = sqrt(fdum);

	    for ( i = 0; i < nx; i++)
	    {
		for ( j = 0; j < ny; j++)
		{

		    bc_type = 0;   /* This is the default "none" boundary condition. */
		    if(i == 0)      bc_type = 19;  /* This is for inflow on the left given by "ifxm". */
		    if(i == nx - 1) bc_type = 14;  /* This is for the exiting on the right. It is the
                                                      2nd free BC given by "frxp".  */
		    if(j == 0)      bc_type = 9;  /* This is for the flow at the bottom of the boundary.
                                                     It is the freestream given by "ffym". */
		    if(j == ny - 1) bc_type = 10;  /* This is for the flow at the top of the boundary.
                                                      It is the freestream given by "ffyp". */

		    memset(text, 0, 10*sizeof(char));

		    check = meshBCset( &bc_type, text);
		    if(!check) printf( " Problems with meshBCset\n");

		    fprintf( o3, "%4d %4d  %10s   ", i, j, text); 

		    fprintf( o3, "%12.4e %12.4e %12.4e %12.4e %12.4e %12.4e %12.4e %12.4e\n",
			dns, temp, prs, vx, vy, enthal, ene, mach);
		}
	    }
	}
	else
	{
	    prs = 0.0; dns = 0.0; temp = 0.0; ene = 0.0; enthal = 0.0;
	    vmag = sqrt(vx*vx + vy*vy);

	    for ( i = 0; i < nx; i++)
	    {
		for ( j = 0; j < ny; j++)
		{

		    bc_type = 0;   /* This is the default "none" boundary condition. */
		    if(i == 0)      bc_type = 19;  /* This is for inflow on the left given by "ifxm". */
		    if(i == nx - 1) bc_type = 32;  /* This is for the exiting on the right. It is the 2nd free BC
		                                      given by "fr2xp".  */
		    if(j == 0)      bc_type = 27;  /* This is for the flow at the bottom of the boundary. It is the
		                                      freestream given by "fsym". */
		    if(j == ny - 1) bc_type = 28;  /* This is for the flow at the top of the boundary.  It is the
		                                      freestream given by "fsyp". */
		    if(i == 0 && j == 0) bc_type = 63;  /* This is the 0 pressure node.  It is given by "prs0". */

		    memset(text, 0, 10*sizeof(char));

		    check = meshBCset( &bc_type, text);
		    if(!check) printf( " Problems with meshBCset\n");

		    fprintf( o3, "%4d %4d  %10s   ", i, j, text); 

		    fprintf( o3, "%12.4e %12.4e %12.4e %12.4e\n",
		        prs, vx, vy, vmag);

		}
	    }
	}

	fprintf( o3, "%4d ",-10);

	return 1;
}

