/* The functions in this file handle the regridding of the data for the subgrids. 
   This consists of 3 operations:

   1) Calculating the fine grid discretizations for dx, dy, and dt as well as the
      subgrid's dimensions.

   2) regridding from the coarse grid to the fine grid

   3) regridding from fine grid to the coarse after calculation on the refined grid is made.

   This is the truest regridding scheme although it is not as stable as that found in
   R.

                        Updated 4/9/08

    SLFCFD source file
    Version:  1.1
    Copyright (C) 2007  San Le

    The source code contained in this file is released under the
    terms of the GNU Library General Public License.
 */

#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "../compressible/cfdconst.h"
#include "../compressible/cfdstruct.h"
#include "../AMR/amr.h"

extern double dx, dy, dt;

int setSubgrid( SUBGRID *subgrid_f )
{
/* This code does 2 things:

   1) Sets the value of dx, dy, and dt for the subgrid_f.

   2) Sets the dimensions of the subgrid_f which doubles the number of points in i and j
      as the original grid it is in.

                        Updated 3/28/07
*/

	int i, i2, j, j2, node_num, dum, nx_fine, ny_fine, ny_coarse;
	int ce, ce2, xm1, xm2, xp1, ym1, ym2, yp1, yp2, xp1yp1;

	subgrid_f[0].nx = 2.0*(subgrid_f[0].iend - subgrid_f[0].istart) + 1;
	subgrid_f[0].ny = 2.0*(subgrid_f[0].jend - subgrid_f[0].jstart) + 1;

	switch(subgrid_f[0].refine_num)
	{
	    case 1: /* Divide grid discretization by 2 */
		subgrid_f[0].dx = dx/2.0;
		subgrid_f[0].dy = dy/2.0;
		subgrid_f[0].dt = dt/2.0;
	    break;
	    case 2: /* Divide grid discretization by 4 */
		subgrid_f[0].dx = dx/4.0;
		subgrid_f[0].dy = dy/4.0;
		subgrid_f[0].dt = dt/4.0;
	    break;
	    case 3: /* Divide grid discretization by 8 */
		subgrid_f[0].dx = dx/8.0;
		subgrid_f[0].dy = dy/8.0;
		subgrid_f[0].dt = dt/8.0;
	    break;
	    default:
		subgrid_f[0].dx = dx;
		subgrid_f[0].dy = dy;
		subgrid_f[0].dt = dt;
	    break;
	}

	return 1;
}

int refineGridSet_AMR( double *u_c, double *v_c, double *prs_c, double *u_f, double *v_f, double *prs_f,
	SUBGRID subgrid_c, SUBGRID subgrid_f )
{
/* This function prepares the subgrid.  The resulting subgrid may be a
   subgrid of the intial mesh or a subgrid of a subgrid.  This code does
   essentially 2 things:

   1) It takes the patch of grid and doubles the number nodal points in
      the i and j direction. 

   2) The nodal values from the coarser grid are assiged to the corresponding
      points of the subgrid.  Then the points which fall in between the nodes
      of the coarser grid have their values interpolted from the original.

                        Updated 4/9/07
*/

	int i, i2, j, j2, node_num, dum, nx_fine, ny_fine, ny_coarse;
	int ce, ce2, xm1, xm2, xp1, ym1, ym2, yp1, yp2, xp1yp1;

	nx_fine = subgrid_f.nx;
	ny_fine = subgrid_f.ny;

	ny_coarse = subgrid_c.ny;

	j2 = 0;
	for( j = subgrid_f.jstart; j < subgrid_f.jend + 1; ++j)
	{
	    i2 = 0;
	    for( i = subgrid_f.istart; i < subgrid_f.iend + 1; ++i)
	    {

/* For the fine nodes which directly fall on the coarse nodes (i,j) */

		ce = ny_coarse*i + j;
		xp1 = ny_coarse*(i+1) + j;
		yp1 = ny_coarse*i + j+1;
		xp1yp1 = ny_coarse*(i+1) + j+1;
		ce2 = ny_fine*i2 + j2;
		u_f[ce2] = u_c[ce];
		v_f[ce2] = v_c[ce];
		prs_f[ce2] = prs_c[ce];
		/*printf("%7d %7d %7d %7d %7d %7d\n", i, j, i2, j2, ce, ce2);*/ 

		if( i < subgrid_f.iend )
		{
/* For the fine nodes which directly fall between the coarse nodes at (i,j) and (i+1,j) */

		    ce2 = ny_fine*(i2+1) + j2;
		    u_f[ce2] = .5*(u_c[ce] + u_c[xp1]);
		    v_f[ce2] = .5*(v_c[ce] + v_c[xp1]);
		    prs_f[ce2] = .5*(prs_c[ce] + prs_c[xp1]);
		}

		if( j < subgrid_f.jend )
		{
/* For the fine nodes which directly fall between the coarse nodes at (i,j) and (i,j+1) */

		    ce2 = ny_fine*i2 + j2+1; 
		    u_f[ce2] = .5*(u_c[ce] + u_c[yp1]);
		    v_f[ce2] = .5*(v_c[ce] + v_c[yp1]);
		    prs_f[ce2] = .5*(prs_c[ce] + prs_c[yp1]);
		}

		if( i < subgrid_f.iend && j < subgrid_f.jend )
		{
/* For the fine nodes which directly fall between the coarse nodes at (i,j), (i+1,j), (i,j+1),
   and (i+1,j+1) */

		    ce2 = ny_fine*(i2+1) + j2+1; 
		    u_f[ce2] = .25*(u_c[ce] + u_c[xp1] + u_c[yp1] + u_c[xp1yp1]);
		    v_f[ce2] = .25*(v_c[ce] + v_c[xp1] + v_c[yp1] + v_c[xp1yp1]);
		    prs_f[ce2] = .25*(prs_c[ce] + prs_c[xp1] + prs_c[yp1] + prs_c[xp1yp1]);
		}

		i2 += 2;
	    }
	    j2 += 2;
	}

	return 1;
}

int refineGridSetBC( double *u_c, double *v_c, double *prs_c, double *u_f, double *v_f, double *prs_f,
	SUBGRID subgrid_c, SUBGRID subgrid_f )
{
/* This function sets the boundary of the subgrid to the values of the corresponding
   nodes on the coarse grid.  The boundary nodes which fall in between the coarse
   grid nodes are interpolated.  This function is used to retain the internal refined
   grid values from the previous calculation.

                        Updated 4/9/07
*/

	int i, i2, j, j2, node_num, dum, nx_fine, ny_fine, ny_coarse;
	int ce, ce2, xm1, xm2, xp1, ym1, ym2, yp1, yp2, xp1yp1;

	nx_fine = subgrid_f.nx;
	ny_fine = subgrid_f.ny;

	ny_coarse = subgrid_c.ny;

/* For the nodes at the top and bottem of the boundary */

	i2 = 0;
	for( i = subgrid_f.istart; i < subgrid_f.iend + 1; ++i)
	{
/* For the nodes at the bottom of the boundary */

		j = subgrid_f.jstart;
		j2 = 0;

/* For the fine nodes which directly fall on the coarse nodes (i,j) */

		ce = ny_coarse*i + j;
		ce2 = ny_fine*i2 + j2;
		u_f[ce2] = u_c[ce];
		v_f[ce2] = v_c[ce];
		prs_f[ce2] = prs_c[ce];
		/*printf("%7d %7d %7d %7d %7d %7d\n", i, j, i2, j2, ce, ce2);*/ 

		if( i < subgrid_f.iend )
		{
/* For the fine nodes which directly fall between the coarse nodes at (i,j) and (i+1,j) */

		    xp1 = ny_coarse*(i+1) + j;
		    ce2 = ny_fine*(i2+1) + j2;
		    u_f[ce2] = .5*(u_c[ce] + u_c[xp1]);
		    v_f[ce2] = .5*(v_c[ce] + v_c[xp1]);
		    prs_f[ce2] = .5*(prs_c[ce] + prs_c[xp1]);
		}

/* For the nodes at the top of the boundary */

		j = subgrid_f.jend;
		j2 = ny_fine - 1;

/* For the fine nodes which directly fall on the coarse nodes (i,j) */

		ce = ny_coarse*i + j;
		ce2 = ny_fine*i2 + j2;
		u_f[ce2] = u_c[ce];
		v_f[ce2] = v_c[ce];
		prs_f[ce2] = prs_c[ce];
		/*printf("%7d %7d %7d %7d %7d %7d\n", i, j, i2, j2, ce, ce2);*/ 

		if( i < subgrid_f.iend )
		{
/* For the fine nodes which directly fall between the coarse nodes at (i,j) and (i+1,j) */

		    xp1 = ny_coarse*(i+1) + j;
		    ce2 = ny_fine*(i2+1) + j2;
		    u_f[ce2] = .5*(u_c[ce] + u_c[xp1]);
		    v_f[ce2] = .5*(v_c[ce] + v_c[xp1]);
		    prs_f[ce2] = .5*(prs_c[ce] + prs_c[xp1]);
		}

		i2 += 2;
	}

/* For the nodes at the left and right of the boundary */

	j2 = 0;
	for( j = subgrid_f.jstart; j < subgrid_f.jend; ++j)
	{
/* For the nodes at the left of the boundary */

		i = subgrid_f.istart;
		i2 = 0;

/* For the fine nodes which directly fall on the coarse nodes (i,j) */

		ce = ny_coarse*i + j;
		ce2 = ny_fine*i2 + j2;
		u_f[ce2] = u_c[ce];
		v_f[ce2] = v_c[ce];
		prs_f[ce2] = prs_c[ce];
		/*printf("%7d %7d %7d %7d %7d %7d\n", i, j, i2, j2, ce, ce2);*/ 

		if( j < subgrid_f.jend )
		{
/* For the fine nodes which directly fall between the coarse nodes at (i,j) and (i,j+1) */

		    yp1 = ny_coarse*i + j+1;
		    ce2 = ny_fine*i2 + j2+1; 
		    u_f[ce2] = .5*(u_c[ce] + u_c[yp1]);
		    v_f[ce2] = .5*(v_c[ce] + v_c[yp1]);
		    prs_f[ce2] = .5*(prs_c[ce] + prs_c[yp1]);
		}

/* For the nodes at the right of the boundary */

		i = subgrid_f.iend;
		i2 = nx_fine - 1;

/* For the fine nodes which directly fall on the coarse nodes (i,j) */

		ce = ny_coarse*i + j;
		ce2 = ny_fine*i2 + j2;
		u_f[ce2] = u_c[ce];
		v_f[ce2] = v_c[ce];
		prs_f[ce2] = prs_c[ce];
		/*printf("%7d %7d %7d %7d %7d %7d\n", i, j, i2, j2, ce, ce2);*/ 

		if( j < subgrid_f.jend )
		{
/* For the fine nodes which directly fall between the coarse nodes at (i,j) and (i,j+1) */

		    yp1 = ny_coarse*i + j+1;
		    ce2 = ny_fine*i2 + j2+1; 
		    u_f[ce2] = .5*(u_c[ce] + u_c[yp1]);
		    v_f[ce2] = .5*(v_c[ce] + v_c[yp1]);
		    prs_f[ce2] = .5*(prs_c[ce] + prs_c[yp1]);
		}

		j2 += 2;
	}

	/*printf("%7d %7d %7d %7d\n",
		subgrid_f.iend - subgrid_f.istart, subgrid_f.jend - subgrid_f.jstart, i2, j2); */

	return 1;
}



int coarseGridSet_AMR( double *u_c, double *v_c, double *prs_c, double *u_f, double *v_f, double *prs_f,
	SUBGRID subgrid_c, SUBGRID subgrid_f )
{
/* This function prepares takes the data values from the fine grid and
   puts it on the corresponding ponits of the coarse grid. 

                        Updated 4/9/07
*/

	int i, i2, j, j2, node_num, dum, nx_fine, ny_fine, ny_coarse;
	int ce, ce2, xm1, xm2, xp1, ym1, ym2, yp1, yp2, xp1yp1;

	nx_fine = subgrid_f.nx;
	ny_fine = subgrid_f.ny;

	ny_coarse = subgrid_c.ny;

	j2 = 0;
	for( j = subgrid_f.jstart; j < subgrid_f.jend + 1; ++j)
	{
	    i2 = 0;
	    for( i = subgrid_f.istart; i < subgrid_f.iend + 1; ++i)
	    {

/* For the fine nodes which directly fall on the coarse nodes (i,j) */

		ce = ny_coarse*i + j;
		ce2 = ny_fine*i2 + j2;
		u_c[ce] = u_f[ce2];
		v_c[ce] = v_f[ce2];
		prs_c[ce] = prs_f[ce2];
		/*printf("%7d %7d %7d %7d %7d %7d\n", i, j, i2, j2, ce, ce2);*/ 
		i2 += 2;
	    }
	    j2 += 2;
	}

	return 1;
}


int refineGridSet( double *vec_coarse, double *vec_fine, SUBGRID subgrid_c, SUBGRID subgrid_f )
{
/* This program prepares the subgrid.  The resulting subgrid may be a
   subgrid of the intial mesh or a subgrid of a subgrid.  This code does
   essentially 2 things:

   1) It takes the patch of grid and doubles the number nodal points in
      the i and j direction. 

   2) The nodal values from the coarser grid are assiged to the corresponding
      points of the subgrid.  Then the points which fall in between the nodes
      of the coarser grid have their values interpolted from the original.

                        Updated 4/9/07
*/

	int i, i2, j, j2, node_num, dum, nx_fine, ny_fine, ny_coarse;
	int ce, ce2, xm1, xm2, xp1, ym1, ym2, yp1, yp2, xp1yp1;

	nx_fine = subgrid_f.nx;
	ny_fine = subgrid_f.ny;

	ny_coarse = subgrid_c.ny;

	j2 = 0;
	for( j = subgrid_f.jstart; j < subgrid_f.jend + 1; ++j)
	{
	    i2 = 0;
	    for( i = subgrid_f.istart; i < subgrid_f.iend + 1; ++i)
	    {

/* For the fine nodes which directly fall on the coarse nodes (i,j) */

		ce = ny_coarse*i + j;
		ce = ny_coarse*i + j;
		xp1 = ny_coarse*(i+1) + j;
		yp1 = ny_coarse*i + j+1;
		xp1yp1 = ny_coarse*(i+1) + j+1;
		ce2 = ny_fine*i2 + j2;
		vec_fine[ce2] = vec_coarse[ce];
		/*printf("%7d %7d %7d %7d %7d %7d\n", i, j, i2, j2, ce, ce2);*/

		if( i < subgrid_f.iend )
		{
/* For the fine nodes which directly fall between the coarse nodes at (i,j) and (i+1,j) */

		    ce2 = ny_fine*(i2+1) + j2;
		    vec_fine[ce2] = .5*(vec_coarse[ce] + vec_coarse[xp1]);
		}
		if( j < subgrid_f.jend )
		{
/* For the fine nodes which directly fall between the coarse nodes at (i,j) and (i,j+1) */

		    ce2 = ny_fine*i2 + j2+1;
		    vec_fine[ce2] = .5*(vec_coarse[ce] + vec_coarse[yp1]);
		}

		if( i < subgrid_f.iend && j < subgrid_f.jend )
		{
/* For the fine nodes which directly fall between the coarse nodes at (i,j), (i+1,j), (i,j+1),
   and (i+1,j+1) */

		    ce2 = ny_fine*(i2+1) + j2+1;
		    vec_fine[ce2] = .25*(vec_coarse[ce] + vec_coarse[xp1] + vec_coarse[yp1] + vec_coarse[xp1yp1]);
		}

		i2 += 2;
	    }
	    j2 += 2;
	}

	return 1;
}

